import os
import shutil
import sys
import tempfile
from multiprocessing import Process
from pathlib import Path
from unittest import mock
from unittest.mock import MagicMock

import pytest
import sqlalchemy as sa
from alembic.script import ScriptDirectory
from pytest_mock.plugin import MockerFixture
from typer.testing import CliRunner

from quetz import cli
from quetz.config import Config
from quetz.db_models import Base, Identity, User
from quetz.testing.utils import Interrupt

runner = CliRunner()


@pytest.fixture
def user_group():
    return "admins"


@pytest.fixture
def default_role():
    return None


@pytest.fixture
def config_extra(user_group, default_role):
    config_values = ["[users]"]
    if user_group is not None:
        config_values.append(f'{user_group} = ["github:bartosz"]')
    if default_role:
        config_values.append(f"{default_role=}")
    return "\n".join(config_values)


@pytest.fixture
def user_with_identity(user, db):
    identity = Identity(user=user, provider="github", identity_id="1")
    db.add(identity)
    db.commit()
    return identity


def get_user(db, config_dir, username="bartosz"):
    def get_db(_):
        return db

    with mock.patch("quetz.cli.get_session", get_db):
        cli.add_user_roles(config_dir)

    return db.query(User).filter(User.username == username).one_or_none()


def test_init_db(db, config, config_dir, mocker):
    _run_migrations: MagicMock = mocker.patch("quetz.cli._run_migrations")
    cli.init_db(config_dir)
    _run_migrations.assert_called_once()


@pytest.mark.parametrize(
    "user_group,expected_role",
    [("admins", "owner"), ("maintainers", "maintainer"), ("members", "member")],
)
def test_create_user_from_config(
    db, config, config_dir, user_group, expected_role, mocker, user_with_identity
):
    user = get_user(db, config_dir)
    assert user

    assert user.role == expected_role
    assert user.username == "bartosz"


@pytest.mark.parametrize("user_group", [None])
def test_set_user_roles_no_user(
    db, config, config_dir, user_group, mocker: MockerFixture
):
    user = get_user(db, config_dir)

    assert user is None


def test_set_user_roles_user_exists(
    db, config, config_dir, user, mocker, user_with_identity
):
    user = get_user(db, config_dir)
    assert user

    assert user.role == 'owner'
    assert user.username == "bartosz"


@pytest.mark.parametrize("default_role", [None, "member"])
@pytest.mark.parametrize("current_role", ['owner', 'member', 'maintainer'])
def test_set_user_roles_user_has_role(
    db, config, config_dir, user, mocker, user_with_identity, current_role, default_role
):
    user.role = current_role
    db.commit()
    user = get_user(db, config_dir)
    assert user

    # role shouldn't be changed unless it's default role
    if current_role != default_role:
        assert user.role == current_role
    else:
        assert user.role == "owner"
    assert user.username == "bartosz"


@pytest.mark.parametrize("config_extra", ['[users]\nadmins = ["dummy:alice"]\n'])
def test_init_db_create_test_users(db, config, mocker, config_dir):
    _run_migrations: MagicMock = mocker.patch("quetz.cli._run_migrations")

    def get_db(_):
        return db

    with mock.patch("quetz.cli.get_session", get_db):
        cli.create(
            Path(config_dir) / "new-deployment",
            copy_conf="config.toml",
            dev=True,
        )

    user = db.query(User).filter(User.username == "alice").one_or_none()

    assert user.role == "owner"

    _run_migrations.assert_called_once()


@pytest.fixture
def refresh_db(engine, database_url):
    Base.metadata.drop_all(engine)
    try:
        with engine.connect() as connection:
            connection.execute(sa.text("DROP TABLE alembic_version"))
    except sa.exc.DatabaseError:
        pass


def test_run_migrations(
    config, sql_connection, engine, database_url, alembic_config, refresh_db
):
    db = sql_connection
    with pytest.raises(sa.exc.DatabaseError):
        db.execute(sa.text("SELECT * FROM users"))

    cli._run_migrations(alembic_config=alembic_config)

    db.execute(sa.text("SELECT * FROM users"))


def test_make_migrations_quetz(mocker, config, config_dir):
    revision = mocker.patch("alembic.command.revision")

    # new revision for main tree
    cli.make_migrations(
        config_dir, message="test revision", plugin="quetz", initialize=False
    )

    revision.assert_called_with(
        mock.ANY,
        message="test revision",
        autogenerate=True,
        head="quetz@head",
        version_path=None,
    )


@pytest.fixture(scope="module")
def dummy_migration_plugin() -> Path:
    path = Path(tempfile.mkdtemp(prefix="quetz"))
    plugin_dir = str(path)
    pkg_path = path / "dummy"

    os.makedirs(pkg_path / "versions")
    with open(pkg_path / "dummy_module.py", 'w') as fid:
        fid.write("class DummyPlugin: pass")
    with open(pkg_path / "__init__.py", 'w') as fid:
        fid.write("")

    # add entry points
    os.makedirs(path / "dummy-0.0.0.dist-info")
    with open(path / "dummy-0.0.0.dist-info" / "entry_points.txt", 'w') as fid:
        fid.writelines(
            [
                "[quetz.models]\nquetz-plugin = dummy.dummy_module\n",
                "[quetz.migrations]\nquetz-plugin = dummy",
            ]
        )

    sys.path.insert(0, plugin_dir)

    yield pkg_path

    shutil.rmtree(path, ignore_errors=True)
    sys.path.remove(plugin_dir)


def test_make_migrations_plugin(mocker, config, config_dir, dummy_migration_plugin):
    revision = mocker.patch("alembic.command.revision")

    # initialize a plugin
    cli.make_migrations(
        config_dir, message="test revision", plugin="quetz-plugin", initialize=True
    )

    version_path = os.path.join(dummy_migration_plugin, "versions")
    revision.assert_called_with(
        mock.ANY,
        message="test revision",
        autogenerate=True,
        head="base",
        depends_on="quetz",
        version_path=version_path,
        branch_label="quetz-plugin",
        splice=True,
    )

    # add revision to a plugin
    cli.make_migrations(
        config_dir, message="revision v2", plugin="quetz-plugin", initialize=False
    )
    revision.assert_called_with(
        mock.ANY,
        message="revision v2",
        autogenerate=True,
        head="quetz-plugin@head",
        version_path=version_path,
    )


def test_make_migrations_plugin_with_alembic(
    config, config_dir, dummy_migration_plugin: Path, alembic_config, engine
):
    # make sure db is up-to-date
    cli._run_migrations(alembic_config=alembic_config)

    alembic_config.set_main_option(
        "version_locations",
        os.path.join(dummy_migration_plugin, "versions") + " quetz:migrations/versions",
    )

    # initialize a plugin
    cli._make_migrations(
        None,
        message="test revision",
        plugin_name="quetz-plugin",
        initialize=True,
        alembic_config=alembic_config,
    )

    migration_scripts = list(
        (dummy_migration_plugin / "versions").glob("*_test_revision.py")
    )
    assert migration_scripts

    # apply migrations
    cli._run_migrations(alembic_config=alembic_config)

    # add a new revision

    class TestPluginModel(Base):
        __tablename__ = "test_plugin_table"
        pk = sa.Column(sa.String, primary_key=True)

    cli._make_migrations(
        None,
        message="new revision",
        plugin_name="quetz-plugin",
        alembic_config=alembic_config,
    )

    migration_scripts = list(
        (dummy_migration_plugin / "versions").glob("*_new_revision.py")
    )
    assert migration_scripts

    with open(migration_scripts[0]) as fid:
        content = fid.read()

    assert "test_plugin_table" in content

    # clean up

    for p in (dummy_migration_plugin / "versions").glob("*.py"):
        os.remove(p)

    Base.metadata.drop_all(engine)
    try:
        with engine.connect() as connection:
            connection.execute(sa.text("DROP TABLE alembic_version"))
    except sa.exc.DatabaseError:
        pass


alembic_env = """
from alembic import context

config = context.config

from sqlalchemy import MetaData
target_metadata = MetaData()

connection = config.attributes.get('connection')

context.configure(connection=connection, target_metadata=target_metadata)
with context.begin_transaction():
    context.run_migrations()
"""

script_mako = """
revision = ${repr(up_revision)}
down_revision = ${repr(down_revision)}
branch_labels = ${repr(branch_labels)}
depends_on = ${repr(depends_on)}

def upgrade(): pass

def downgrade(): pass
"""

quetz_rev = """
revision = "0000"
down_revision = None
branch_labels = ("quetz", )
depends_on = None

def upgrade(): pass

def downgrade(): pass
"""


def test_multi_head(
    config, config_dir, dummy_migration_plugin: Path, alembic_config, engine, refresh_db
):
    quetz_migrations_path = Path(config_dir) / "migrations"
    quetz_versions_path = quetz_migrations_path / "versions"

    alembic_config.config_file_name = quetz_migrations_path / "alembic.ini"

    os.makedirs(quetz_versions_path)

    plugin_versions_path = Path(dummy_migration_plugin) / "versions"

    with open(quetz_migrations_path / "env.py", "w") as fid:
        fid.write(alembic_env)
    with open(quetz_migrations_path / "script.py.mako", "w") as fid:
        fid.write(script_mako)
    with open(quetz_versions_path / "0000_initial.py", 'w') as fid:
        fid.write(quetz_rev)

    alembic_config.set_main_option(
        "version_locations",
        " ".join(map(str, [plugin_versions_path, quetz_versions_path])),
    )
    alembic_config.set_main_option("script_location", str(quetz_migrations_path))
    cli._run_migrations(alembic_config=alembic_config)

    # initialize a plugin
    cli._make_migrations(
        None,
        message="test revision",
        plugin_name="quetz-plugin",
        initialize=True,
        alembic_config=alembic_config,
    )
    cli._run_migrations(alembic_config=alembic_config)

    rev_file = next((plugin_versions_path).glob("*test_revision.py"))
    with open(rev_file) as fid:
        content = fid.read()
    assert 'down_revision = None' in content
    assert "depends_on = 'quetz'" in content
    import re

    m = re.search("revision = '(.*)'", content)
    assert m
    plugin_rev_1 = m.groups()[0]

    # second revision quetz
    cli._make_migrations(
        None,
        message="test revision",
        alembic_config=alembic_config,
    )
    cli._run_migrations(alembic_config=alembic_config)

    rev_file = next((quetz_versions_path).glob("*test_revision.py"))
    with open(rev_file) as fid:
        content = fid.read()
    assert "down_revision = '0000'" in content

    # second revision plugin
    cli._make_migrations(
        None,
        message="plugin rev 2",
        plugin_name="quetz-plugin",
        alembic_config=alembic_config,
    )
    rev_file = next((plugin_versions_path).glob("*plugin_rev_2.py"))

    with open(rev_file) as fid:
        content = fid.read()
    assert f"down_revision = '{plugin_rev_1}'" in content

    cli._run_migrations(alembic_config=alembic_config)

    # check heads

    script_directory = ScriptDirectory.from_config(alembic_config)
    heads = script_directory.get_revisions("heads")
    assert len(heads) == 2

    for p in (plugin_versions_path).glob("*.py"):
        os.remove(p)

    try:
        with engine.connect() as connection:
            connection.execute(sa.text("DROP TABLE alembic_version"))
    except sa.exc.DatabaseError:
        pass


def test_create_exists_ok():
    """Nothing happens if --exists-ok is enabled."""
    with mock.patch("quetz.cli._is_deployment", lambda x: True):
        res = runner.invoke(cli.app, ["create", "test", "--exists-ok"])
        assert res.exit_code == 0


wrong_cli_args = [
    ["create", "test"],
    ["create", "test", "--copy-conf", "config.toml"],
    ["create", "test", "--create-conf"],
    ["create", "test", "--copy-conf", "config.toml", "--create-conf"],
    ["create", "test", "--delete"],
]


@pytest.mark.parametrize('cli_args', wrong_cli_args)
def test_create_exists_errors(cli_args):
    """Create command raises if deployment exists and not force deleted."""
    with mock.patch("quetz.cli._is_deployment", lambda x: True):
        res = runner.invoke(cli.app, cli_args)
        assert res.exit_code == 1
        assert (
            "Use the start command to start a deployment or specify"
            " --delete with --copy-conf or --create-conf." in res.output
        )
        assert "Aborted" in res.output


@pytest.fixture()
def empty_deployment_dir() -> Path:
    path = Path(tempfile.mkdtemp()).resolve()
    try:
        yield path
    finally:
        shutil.rmtree(path, ignore_errors=True)


@pytest.fixture()
def empty_config_on_exit() -> None:
    yield
    try:
        del os.environ["QUETZ_CONFIG_FILE"]
    except KeyError:
        pass
    Config._instances = {}


def test_create_conf(empty_deployment_dir: Path, empty_config_on_exit: None):
    """Create command with create conf cretes the needed files and folder."""
    runner.invoke(cli.app, ['create', str(empty_deployment_dir), '--create-conf'])
    assert empty_deployment_dir.joinpath('config.toml').exists()
    assert empty_deployment_dir.joinpath('channels').exists()
    assert empty_deployment_dir.joinpath('quetz.sqlite').exists()


def test_create_exists_delete(empty_deployment_dir: Path, empty_config_on_exit: None):
    """Existing deployment is removed if delete arg is used."""
    runner.invoke(cli.app, ['create', str(empty_deployment_dir), '--create-conf'])
    creation_time = empty_deployment_dir.stat().st_mtime
    runner.invoke(
        cli.app, ['create', str(empty_deployment_dir), '--delete', '--create-conf']
    )
    assert empty_deployment_dir.stat().st_mtime > creation_time
    assert empty_deployment_dir.joinpath('config.toml').exists()
    assert empty_deployment_dir.joinpath('channels').exists()
    assert empty_deployment_dir.joinpath('quetz.sqlite').exists()


def test_create_no_config(empty_deployment_dir: Path, empty_config_on_exit: None):
    """Error on create command with empty deployment and no config."""
    res = runner.invoke(cli.app, ['create', str(empty_deployment_dir)])
    assert res.exit_code == 1
    assert "No configuration file provided." in res.output


def test_create_extra_file_in_deployment(
    empty_deployment_dir: Path, empty_config_on_exit: None
):
    """Error on create command with extra files in deployment."""
    empty_deployment_dir.joinpath('extra_file.txt').touch()
    res = runner.invoke(cli.app, ['create', str(empty_deployment_dir)])
    assert res.exit_code == 1
    assert "Quetz deployment not allowed at" in res.output


def test_create_missing_copy_conf(
    empty_deployment_dir: Path, empty_config_on_exit: None
):
    """Error on create command with wrong copy-conf path."""
    res = runner.invoke(
        cli.app, ['create', str(empty_deployment_dir), "--copy-conf", "none.toml"]
    )
    assert res.exit_code == 1
    assert "Config file to copy does not exist" in res.output


@pytest.fixture()
def create_channels_dir(config_dir) -> None:
    os.mkdir(Path(config_dir).resolve() / "channels")
    yield


@pytest.fixture()
def session_secret_environment_variable() -> None:
    os.environ["QUETZ_SESSION_SECRET"] = "test"
    yield
    if "QUETZ_SESSION_SECRET" in os.environ:
        del os.environ["QUETZ_SESSION_SECRET"]


@pytest.fixture()
def database_url_environment_variable(database_url) -> None:
    os.environ["QUETZ_SQLALCHEMY_DATABASE_URL"] = database_url
    yield
    if "QUETZ_SQLALCHEMY_DATABASE_URL" in os.environ:
        del os.environ["QUETZ_SQLALCHEMY_DATABASE_URL"]


@pytest.fixture()
def s3_environment_variable() -> None:
    os.environ["QUETZ_S3_ACCESS_KEY"] = "fake_key"
    yield
    if "QUETZ_S3_ACCESS_KEY" in os.environ:
        del os.environ["QUETZ_S3_ACCESS_KEY"]


@pytest.fixture()
def mandatory_environment_variables(
    database_url_environment_variable: None, session_secret_environment_variable: None
) -> None:
    yield


@pytest.mark.timeout(1)
def test_start_without_database_url(
    empty_config_on_exit: None, session_secret_environment_variable: None
):
    """Error starting server without database url"""
    res = runner.invoke(cli.app, ["start"])
    assert res.exit_code == 1
    assert "'database_url' unset" in str(res.exception)


@pytest.mark.timeout(1)
def test_start_without_session_secret(
    empty_config_on_exit: None, database_url_environment_variable: None
):
    """Error starting server without session secret"""
    res = runner.invoke(cli.app, ["start"])
    assert res.exit_code == 1
    assert "'secret' unset" in str(res.exception)


@pytest.mark.timeout(1)
def test_start_server_local_without_deployment(
    empty_config_on_exit: None, mandatory_environment_variables: None
):
    """Error starting server without deployment directory"""

    res = runner.invoke(cli.app, ["start"])
    assert res.exit_code == 1
    assert "The specified directory is not a deployment" in res.output


@pytest.mark.parametrize("sqlite_in_memory", [False])
@pytest.mark.timeout(1)
def test_start_server_local_with_deployment_and_config_file(
    empty_config_on_exit: None, config, config_dir, create_channels_dir, create_tables
):
    """Starting server with deployment directory"""

    p = Process(target=cli.app, args=(["start", config_dir, "--port", "8001"],))
    with Interrupt():
        p.start()
    p.join()

    assert p.exitcode == 0


@pytest.mark.parametrize("sqlite_in_memory", [False])
@pytest.mark.timeout(1)
def test_start_server_local_with_deployment_without_config_file(
    empty_config_on_exit: None,
    config_dir,
    create_channels_dir,
    create_tables,
    mandatory_environment_variables: None,
):
    """
    Starting server with deployment directory but no config file,
    using environmental variables instead
    """

    p = Process(target=cli.app, args=(["start", config_dir, "--port", "8001"],))
    with Interrupt():
        p.start()
    p.join()

    assert p.exitcode == 0


@pytest.mark.parametrize("sqlite_in_memory", [False])
@pytest.mark.timeout(1)
def test_start_server_s3_without_deployment_without_config_file(
    empty_config_on_exit: None,
    create_tables,
    mandatory_environment_variables: None,
    s3_environment_variable: None,
):
    """
    Starting server without deployment directory and no config file,
    using environmental variables and remote storage.
    """

    p = Process(target=cli.app, args=(["start", "--port", "8001"],))
    with Interrupt():
        p.start()
    p.join()

    assert p.exitcode == 0



# flake8: noqa
# Configuration file for jupyterhub.

# ------------------------------------------------------------------------------
# Application(SingletonConfigurable) configuration
# ------------------------------------------------------------------------------
## This is an application.

## The date format used by logging formatters for %(asctime)s
#  Default: '%Y-%m-%d %H:%M:%S'
# c.Application.log_datefmt = '%Y-%m-%d %H:%M:%S'

## The Logging format template
#  Default: '[%(name)s]%(highlevel)s %(message)s'
# c.Application.log_format = '[%(name)s]%(highlevel)s %(message)s'

## Set the log level by value or name.
#  Choices: any of [0, 10, 20, 30, 40, 50, 'DEBUG', 'INFO', 'WARN', 'ERROR', 'CRITICAL']
#  Default: 30
c.Application.log_level = 'DEBUG'

## Instead of starting the Application, dump configuration to stdout
#  Default: False
# c.Application.show_config = False

## Instead of starting the Application, dump configuration to stdout (as JSON)
#  Default: False
# c.Application.show_config_json = False

# ------------------------------------------------------------------------------
# JupyterHub(Application) configuration
# ------------------------------------------------------------------------------
## An Application for starting a Multi-User Jupyter Notebook server.

## Maximum number of concurrent servers that can be active at a time.
#
#  Setting this can limit the total resources your users can consume.
#
#  An active server is any server that's not fully stopped. It is considered
#  active from the time it has been requested until the time that it has
#  completely stopped.
#
#  If this many user servers are active, users will not be able to launch new
#  servers until a server is shutdown. Spawn requests will be rejected with a 429
#  error asking them to try again.
#
#  If set to 0, no limit is enforced.
#  Default: 0
# c.JupyterHub.active_server_limit = 0

## Duration (in seconds) to determine the number of active users.
#  Default: 1800
# c.JupyterHub.active_user_window = 1800

## Resolution (in seconds) for updating activity
#
#  If activity is registered that is less than activity_resolution seconds more
#  recent than the current value, the new value will be ignored.
#
#  This avoids too many writes to the Hub database.
#  Default: 30
# c.JupyterHub.activity_resolution = 30

## Grant admin users permission to access single-user servers.
#
#  Users should be properly informed if this is enabled.
#  Default: False
# c.JupyterHub.admin_access = False

## DEPRECATED since version 0.7.2, use Authenticator.admin_users instead.
#  Default: set()
# c.JupyterHub.admin_users = set()

## Allow named single-user servers per user
#  Default: False
# c.JupyterHub.allow_named_servers = False

## Answer yes to any questions (e.g. confirm overwrite)
#  Default: False
# c.JupyterHub.answer_yes = False

## PENDING DEPRECATION: consider using services
#
#  Dict of token:username to be loaded into the database.
#
#  Allows ahead-of-time generation of API tokens for use by externally managed
#  services, which authenticate as JupyterHub users.
#
#  Consider using services for general services that talk to the JupyterHub API.
#  Default: {}
# c.JupyterHub.api_tokens = {}

## Authentication for prometheus metrics
#  Default: True
# c.JupyterHub.authenticate_prometheus = True

## Class for authenticating users.
#
#          This should be a subclass of :class:`jupyterhub.auth.Authenticator`
#
#          with an :meth:`authenticate` method that:
#
#          - is a coroutine (asyncio or tornado)
#          - returns username on success, None on failure
#          - takes two arguments: (handler, data),
#            where `handler` is the calling web.RequestHandler,
#            and `data` is the POST form data from the login page.
#
#          .. versionchanged:: 1.0
#              authenticators may be registered via entry points,
#              e.g. `c.JupyterHub.authenticator_class = 'pam'`
#
#  Currently installed:
#    - default: jupyterhub.auth.PAMAuthenticator
#    - dummy: jupyterhub.auth.DummyAuthenticator
#    - pam: jupyterhub.auth.PAMAuthenticator
#  Default: 'jupyterhub.auth.PAMAuthenticator'
# c.JupyterHub.authenticator_class = 'jupyterhub.auth.PAMAuthenticator'

## The base URL of the entire application.
#
#  Add this to the beginning of all JupyterHub URLs. Use base_url to run
#  JupyterHub within an existing website.
#
#  .. deprecated: 0.9
#      Use JupyterHub.bind_url
#  Default: '/'
# c.JupyterHub.base_url = '/'

## The public facing URL of the whole JupyterHub application.
#
#  This is the address on which the proxy will bind. Sets protocol, ip, base_url
#  Default: 'http://:8000'
# c.JupyterHub.bind_url = 'http://:8000'

## Whether to shutdown the proxy when the Hub shuts down.
#
#  Disable if you want to be able to teardown the Hub while leaving the proxy
#  running.
#
#  Only valid if the proxy was starting by the Hub process.
#
#  If both this and cleanup_servers are False, sending SIGINT to the Hub will
#  only shutdown the Hub, leaving everything else running.
#
#  The Hub should be able to resume from database state.
#  Default: True
# c.JupyterHub.cleanup_proxy = True

## Whether to shutdown single-user servers when the Hub shuts down.
#
#  Disable if you want to be able to teardown the Hub while leaving the single-
#  user servers running.
#
#  If both this and cleanup_proxy are False, sending SIGINT to the Hub will only
#  shutdown the Hub, leaving everything else running.
#
#  The Hub should be able to resume from database state.
#  Default: True
# c.JupyterHub.cleanup_servers = True

## Maximum number of concurrent users that can be spawning at a time.
#
#  Spawning lots of servers at the same time can cause performance problems for
#  the Hub or the underlying spawning system. Set this limit to prevent bursts of
#  logins from attempting to spawn too many servers at the same time.
#
#  This does not limit the number of total running servers. See
#  active_server_limit for that.
#
#  If more than this many users attempt to spawn at a time, their requests will
#  be rejected with a 429 error asking them to try again. Users will have to wait
#  for some of the spawning services to finish starting before they can start
#  their own.
#
#  If set to 0, no limit is enforced.
#  Default: 100
# c.JupyterHub.concurrent_spawn_limit = 100

## The config file to load
#  Default: 'jupyterhub_config.py'
# c.JupyterHub.config_file = 'jupyterhub_config.py'

## DEPRECATED: does nothing
#  Default: False
# c.JupyterHub.confirm_no_ssl = False

## Number of days for a login cookie to be valid. Default is two weeks.
#  Default: 14
# c.JupyterHub.cookie_max_age_days = 14

## The cookie secret to use to encrypt cookies.
#
#  Loaded from the JPY_COOKIE_SECRET env variable by default.
#
#  Should be exactly 256 bits (32 bytes).
#  Default: b''
# c.JupyterHub.cookie_secret = b''

## File in which to store the cookie secret.
#  Default: 'jupyterhub_cookie_secret'
# c.JupyterHub.cookie_secret_file = 'jupyterhub_cookie_secret'

## The location of jupyterhub data files (e.g. /usr/local/share/jupyterhub)
#  Default: '/usr/local/share/jupyterhub'
# c.JupyterHub.data_files_path = '/usr/local/share/jupyterhub'

## Include any kwargs to pass to the database connection. See
#  sqlalchemy.create_engine for details.
#  Default: {}
# c.JupyterHub.db_kwargs = {}

## url for the database. e.g. `sqlite:///jupyterhub.sqlite`
#  Default: 'sqlite:///jupyterhub.sqlite'

# c.JupyterHub.db_url = 'postgresql:///postgres@database:5932/jupyterhub'

## log all database transactions. This has A LOT of output
#  Default: False
# c.JupyterHub.debug_db = False

## DEPRECATED since version 0.8: Use ConfigurableHTTPProxy.debug
#  Default: False
# c.JupyterHub.debug_proxy = False

## If named servers are enabled, default name of server to spawn or open, e.g. by
#  user-redirect.
#  Default: ''
# c.JupyterHub.default_server_name = ''

## The default URL for users when they arrive (e.g. when user directs to "/")
#
#  By default, redirects users to their own server.
#
#  Can be a Unicode string (e.g. '/hub/home') or a callable based on the handler
#  object:
#
#  ::
#
#      def default_url_fn(handler):
#          user = handler.current_user
#          if user and user.admin:
#              return '/hub/admin'
#          return '/hub/home'
#
#      c.JupyterHub.default_url = default_url_fn
#  Default: traitlets.Undefined
# c.JupyterHub.default_url = traitlets.Undefined

## Dict authority:dict(files). Specify the key, cert, and/or ca file for an
#  authority. This is useful for externally managed proxies that wish to use
#  internal_ssl.
#
#  The files dict has this format (you must specify at least a cert)::
#
#      {
#          'key': '/path/to/key.key',
#          'cert': '/path/to/cert.crt',
#          'ca': '/path/to/ca.crt'
#      }
#
#  The authorities you can override: 'hub-ca', 'notebooks-ca', 'proxy-api-ca',
#  'proxy-client-ca', and 'services-ca'.
#
#  Use with internal_ssl
#  Default: {}
# c.JupyterHub.external_ssl_authorities = {}

## Register extra tornado Handlers for jupyterhub.
#
#  Should be of the form ``("<regex>", Handler)``
#
#  The Hub prefix will be added, so `/my-page` will be served at `/hub/my-page`.
#  Default: []
# c.JupyterHub.extra_handlers = []

## DEPRECATED: use output redirection instead, e.g.
#
#  jupyterhub &>> /var/log/jupyterhub.log
#  Default: ''
# c.JupyterHub.extra_log_file = ''

## Extra log handlers to set on JupyterHub logger
#  Default: []
# c.JupyterHub.extra_log_handlers = []

## Generate certs used for internal ssl
#  Default: False
# c.JupyterHub.generate_certs = False

## Generate default config file
#  Default: False
# c.JupyterHub.generate_config = False

## The URL on which the Hub will listen. This is a private URL for internal
#  communication. Typically set in combination with hub_connect_url. If a unix
#  socket, hub_connect_url **must** also be set.
#
#  For example:
#
#      "http://127.0.0.1:8081"
#      "unix+http://%2Fsrv%2Fjupyterhub%2Fjupyterhub.sock"
#
#  .. versionadded:: 0.9
#  Default: ''
# c.JupyterHub.hub_bind_url = ''

## The ip or hostname for proxies and spawners to use for connecting to the Hub.
#
#  Use when the bind address (`hub_ip`) is 0.0.0.0, :: or otherwise different
#  from the connect address.
#
#  Default: when `hub_ip` is 0.0.0.0 or ::, use `socket.gethostname()`, otherwise
#  use `hub_ip`.
#
#  Note: Some spawners or proxy implementations might not support hostnames.
#  Check your spawner or proxy documentation to see if they have extra
#  requirements.
#
#  .. versionadded:: 0.8
#  Default: ''
# c.JupyterHub.hub_connect_ip = ''

## DEPRECATED
#
#  Use hub_connect_url
#
#  .. versionadded:: 0.8
#
#  .. deprecated:: 0.9
#      Use hub_connect_url
#  Default: 0
# c.JupyterHub.hub_connect_port = 0

## The URL for connecting to the Hub. Spawners, services, and the proxy will use
#  this URL to talk to the Hub.
#
#  Only needs to be specified if the default hub URL is not connectable (e.g.
#  using a unix+http:// bind url).
#
#  .. seealso::
#      JupyterHub.hub_connect_ip
#      JupyterHub.hub_bind_url
#
#  .. versionadded:: 0.9
#  Default: ''
# c.JupyterHub.hub_connect_url = ''

## The ip address for the Hub process to *bind* to.
#
#  By default, the hub listens on localhost only. This address must be accessible
#  from the proxy and user servers. You may need to set this to a public ip or ''
#  for all interfaces if the proxy or user servers are in containers or on a
#  different host.
#
#  See `hub_connect_ip` for cases where the bind and connect address should
#  differ, or `hub_bind_url` for setting the full bind URL.
#  Default: '127.0.0.1'
# c.JupyterHub.hub_ip = '127.0.0.1'

## The internal port for the Hub process.
#
#  This is the internal port of the hub itself. It should never be accessed
#  directly. See JupyterHub.port for the public port to use when accessing
#  jupyterhub. It is rare that this port should be set except in cases of port
#  conflict.
#
#  See also `hub_ip` for the ip and `hub_bind_url` for setting the full bind URL.
#  Default: 8081
# c.JupyterHub.hub_port = 8081

## Trigger implicit spawns after this many seconds.
#
#  When a user visits a URL for a server that's not running, they are shown a
#  page indicating that the requested server is not running with a button to
#  spawn the server.
#
#  Setting this to a positive value will redirect the user after this many
#  seconds, effectively clicking this button automatically for the users,
#  automatically beginning the spawn process.
#
#  Warning: this can result in errors and surprising behavior when sharing access
#  URLs to actual servers, since the wrong server is likely to be started.
#  Default: 0
# c.JupyterHub.implicit_spawn_seconds = 0

## Timeout (in seconds) to wait for spawners to initialize
#
#  Checking if spawners are healthy can take a long time if many spawners are
#  active at hub start time.
#
#  If it takes longer than this timeout to check, init_spawner will be left to
#  complete in the background and the http server is allowed to start.
#
#  A timeout of -1 means wait forever, which can mean a slow startup of the Hub
#  but ensures that the Hub is fully consistent by the time it starts responding
#  to requests. This matches the behavior of jupyterhub 1.0.
#
#  .. versionadded: 1.1.0
#  Default: 10
# c.JupyterHub.init_spawners_timeout = 10

## The location to store certificates automatically created by JupyterHub.
#
#  Use with internal_ssl
#  Default: 'internal-ssl'
# c.JupyterHub.internal_certs_location = 'internal-ssl'

## Enable SSL for all internal communication
#
#  This enables end-to-end encryption between all JupyterHub components.
#  JupyterHub will automatically create the necessary certificate authority and
#  sign notebook certificates as they're created.
#  Default: False
# c.JupyterHub.internal_ssl = False

## The public facing ip of the whole JupyterHub application (specifically
#  referred to as the proxy).
#
#  This is the address on which the proxy will listen. The default is to listen
#  on all interfaces. This is the only address through which JupyterHub should be
#  accessed by users.
#
#  .. deprecated: 0.9
#      Use JupyterHub.bind_url
#  Default: ''
# c.JupyterHub.ip = ''

## Supply extra arguments that will be passed to Jinja environment.
#  Default: {}
# c.JupyterHub.jinja_environment_options = {}

## Interval (in seconds) at which to update last-activity timestamps.
#  Default: 300
# c.JupyterHub.last_activity_interval = 300

## Dict of 'group': ['usernames'] to load at startup.
#
#  This strictly *adds* groups and users to groups.
#
#  Loading one set of groups, then starting JupyterHub again with a different set
#  will not remove users or groups from previous launches. That must be done
#  through the API.
#  Default: {}
# c.JupyterHub.load_groups = {}

## The date format used by logging formatters for %(asctime)s
#  See also: Application.log_datefmt
# c.JupyterHub.log_datefmt = '%Y-%m-%d %H:%M:%S'

## The Logging format template
#  See also: Application.log_format
# c.JupyterHub.log_format = '[%(name)s]%(highlevel)s %(message)s'

## Set the log level by value or name.
#  See also: Application.log_level
c.JupyterHub.log_level = 'DEBUG'

## Specify path to a logo image to override the Jupyter logo in the banner.
#  Default: ''
# c.JupyterHub.logo_file = ''

## Maximum number of concurrent named servers that can be created by a user at a
#  time.
#
#  Setting this can limit the total resources a user can consume.
#
#  If set to 0, no limit is enforced.
#  Default: 0
# c.JupyterHub.named_server_limit_per_user = 0

## File to write PID Useful for daemonizing JupyterHub.
#  Default: ''
# c.JupyterHub.pid_file = ''

## The public facing port of the proxy.
#
#  This is the port on which the proxy will listen. This is the only port through
#  which JupyterHub should be accessed by users.
#
#  .. deprecated: 0.9
#      Use JupyterHub.bind_url
#  Default: 8000
# c.JupyterHub.port = 8000

## DEPRECATED since version 0.8 : Use ConfigurableHTTPProxy.api_url
#  Default: ''
# c.JupyterHub.proxy_api_ip = ''

## DEPRECATED since version 0.8 : Use ConfigurableHTTPProxy.api_url
#  Default: 0
# c.JupyterHub.proxy_api_port = 0

## DEPRECATED since version 0.8: Use ConfigurableHTTPProxy.auth_token
#  Default: ''
# c.JupyterHub.proxy_auth_token = ''

## Interval (in seconds) at which to check if the proxy is running.
#  Default: 30
# c.JupyterHub.proxy_check_interval = 30

## The class to use for configuring the JupyterHub proxy.
#
#          Should be a subclass of :class:`jupyterhub.proxy.Proxy`.
#
#          .. versionchanged:: 1.0
#              proxies may be registered via entry points,
#              e.g. `c.JupyterHub.proxy_class = 'traefik'`
#
#  Currently installed:
#    - configurable-http-proxy: jupyterhub.proxy.ConfigurableHTTPProxy
#    - default: jupyterhub.proxy.ConfigurableHTTPProxy
#  Default: 'jupyterhub.proxy.ConfigurableHTTPProxy'
# c.JupyterHub.proxy_class = 'jupyterhub.proxy.ConfigurableHTTPProxy'

## DEPRECATED since version 0.8. Use ConfigurableHTTPProxy.command
#  Default: []
# c.JupyterHub.proxy_cmd = []

## Recreate all certificates used within JupyterHub on restart.
#
#  Note: enabling this feature requires restarting all notebook servers.
#
#  Use with internal_ssl
#  Default: False
# c.JupyterHub.recreate_internal_certs = False

## Redirect user to server (if running), instead of control panel.
#  Default: True
# c.JupyterHub.redirect_to_server = True

## Purge and reset the database.
#  Default: False
# c.JupyterHub.reset_db = False

## Interval (in seconds) at which to check connectivity of services with web
#  endpoints.
#  Default: 60
# c.JupyterHub.service_check_interval = 60

## Dict of token:servicename to be loaded into the database.
#
#  Allows ahead-of-time generation of API tokens for use by externally managed
#  services.
#  Default: {}
# c.JupyterHub.service_tokens = {}

## List of service specification dictionaries.
#
#  A service
#
#  For instance::
#
#      services = [
#          {
#              'name': 'cull_idle',
#              'command': ['/path/to/cull_idle_servers.py'],
#          },
#          {
#              'name': 'formgrader',
#              'url': 'http://127.0.0.1:1234',
#              'api_token': 'super-secret',
#              'environment':
#          }
#      ]
#  Default: []
# c.JupyterHub.services = []

c.JupyterHub.services = [
    {
        'name': 'quetz',
        'url': 'http://web:8000',
        # any secret >8 characters, you'll use api_token to
        # authenticate api requests to the hub from your service
        'oauth_client_id': 'quetz_client',
        'oauth_redirect_uri': 'http://localhost:8000/auth/jupyterhub/authorize',
        'api_token': 'super-secret',
    }
]

## Instead of starting the Application, dump configuration to stdout
#  See also: Application.show_config
# c.JupyterHub.show_config = False

## Instead of starting the Application, dump configuration to stdout (as JSON)
#  See also: Application.show_config_json
# c.JupyterHub.show_config_json = False

## Shuts down all user servers on logout
#  Default: False
# c.JupyterHub.shutdown_on_logout = False

## The class to use for spawning single-user servers.
#
#          Should be a subclass of :class:`jupyterhub.spawner.Spawner`.
#
#          .. versionchanged:: 1.0
#              spawners may be registered via entry points,
#              e.g. `c.JupyterHub.spawner_class = 'localprocess'`
#
#  Currently installed:
#    - default: jupyterhub.spawner.LocalProcessSpawner
#    - localprocess: jupyterhub.spawner.LocalProcessSpawner
#    - simple: jupyterhub.spawner.SimpleLocalProcessSpawner
#  Default: 'jupyterhub.spawner.LocalProcessSpawner'
# c.JupyterHub.spawner_class = 'jupyterhub.spawner.LocalProcessSpawner'

## Path to SSL certificate file for the public facing interface of the proxy
#
#  When setting this, you should also set ssl_key
#  Default: ''
# c.JupyterHub.ssl_cert = ''

## Path to SSL key file for the public facing interface of the proxy
#
#  When setting this, you should also set ssl_cert
#  Default: ''
# c.JupyterHub.ssl_key = ''

## Host to send statsd metrics to. An empty string (the default) disables sending
#  metrics.
#  Default: ''
# c.JupyterHub.statsd_host = ''

## Port on which to send statsd metrics about the hub
#  Default: 8125
# c.JupyterHub.statsd_port = 8125

## Prefix to use for all metrics sent by jupyterhub to statsd
#  Default: 'jupyterhub'
# c.JupyterHub.statsd_prefix = 'jupyterhub'

## Run single-user servers on subdomains of this host.
#
#  This should be the full `https://hub.domain.tld[:port]`.
#
#  Provides additional cross-site protections for javascript served by single-
#  user servers.
#
#  Requires `<username>.hub.domain.tld` to resolve to the same host as
#  `hub.domain.tld`.
#
#  In general, this is most easily achieved with wildcard DNS.
#
#  When using SSL (i.e. always) this also requires a wildcard SSL certificate.
#  Default: ''
# c.JupyterHub.subdomain_host = ''

## Paths to search for jinja templates, before using the default templates.
#  Default: []
# c.JupyterHub.template_paths = []

## Extra variables to be passed into jinja templates
#  Default: {}
# c.JupyterHub.template_vars = {}

## Extra settings overrides to pass to the tornado application.
#  Default: {}
# c.JupyterHub.tornado_settings = {}

## Trust user-provided tokens (via JupyterHub.service_tokens) to have good
#  entropy.
#
#  If you are not inserting additional tokens via configuration file, this flag
#  has no effect.
#
#  In JupyterHub 0.8, internally generated tokens do not pass through additional
#  hashing because the hashing is costly and does not increase the entropy of
#  already-good UUIDs.
#
#  User-provided tokens, on the other hand, are not trusted to have good entropy
#  by default, and are passed through many rounds of hashing to stretch the
#  entropy of the key (i.e. user-provided tokens are treated as passwords instead
#  of random keys). These keys are more costly to check.
#
#  If your inserted tokens are generated by a good-quality mechanism, e.g.
#  `openssl rand -hex 32`, then you can set this flag to True to reduce the cost
#  of checking authentication tokens.
#  Default: False
# c.JupyterHub.trust_user_provided_tokens = False

## Names to include in the subject alternative name.
#
#  These names will be used for server name verification. This is useful if
#  JupyterHub is being run behind a reverse proxy or services using ssl are on
#  different hosts.
#
#  Use with internal_ssl
#  Default: []
# c.JupyterHub.trusted_alt_names = []

## Downstream proxy IP addresses to trust.
#
#  This sets the list of IP addresses that are trusted and skipped when
#  processing the `X-Forwarded-For` header. For example, if an external proxy is
#  used for TLS termination, its IP address should be added to this list to
#  ensure the correct client IP addresses are recorded in the logs instead of the
#  proxy server's IP address.
#  Default: []
# c.JupyterHub.trusted_downstream_ips = []

## Upgrade the database automatically on start.
#
#  Only safe if database is regularly backed up. Only SQLite databases will be
#  backed up to a local file automatically.
#  Default: False
# c.JupyterHub.upgrade_db = False

## Callable to affect behavior of /user-redirect/
#
#  Receives 4 parameters: 1. path - URL path that was provided after /user-
#  redirect/ 2. request - A Tornado HTTPServerRequest representing the current
#  request. 3. user - The currently authenticated user. 4. base_url - The
#  base_url of the current hub, for relative redirects
#
#  It should return the new URL to redirect to, or None to preserve current
#  behavior.
#  Default: None
# c.JupyterHub.user_redirect_hook = None

# ------------------------------------------------------------------------------
# Spawner(LoggingConfigurable) configuration
# ------------------------------------------------------------------------------
## Base class for spawning single-user notebook servers.
#
#  Subclass this, and override the following methods:
#
#  - load_state - get_state - start - stop - poll
#
#  As JupyterHub supports multiple users, an instance of the Spawner subclass is
#  created for each user. If there are 20 JupyterHub users, there will be 20
#  instances of the subclass.

## Extra arguments to be passed to the single-user server.
#
#  Some spawners allow shell-style expansion here, allowing you to use
#  environment variables here. Most, including the default, do not. Consult the
#  documentation for your spawner to verify!
#  Default: []
# c.Spawner.args = []

## An optional hook function that you can implement to pass `auth_state` to the
#  spawner after it has been initialized but before it starts. The `auth_state`
#  dictionary may be set by the `.authenticate()` method of the authenticator.
#  This hook enables you to pass some or all of that information to your spawner.
#
#  Example::
#
#      def userdata_hook(spawner, auth_state):
#          spawner.userdata = auth_state["userdata"]
#
#      c.Spawner.auth_state_hook = userdata_hook
#  Default: None
# c.Spawner.auth_state_hook = None

## The command used for starting the single-user server.
#
#  Provide either a string or a list containing the path to the startup script
#  command. Extra arguments, other than this path, should be provided via `args`.
#
#  This is usually set if you want to start the single-user server in a different
#  python environment (with virtualenv/conda) than JupyterHub itself.
#
#  Some spawners allow shell-style expansion here, allowing you to use
#  environment variables. Most, including the default, do not. Consult the
#  documentation for your spawner to verify!
#  Default: ['jupyterhub-singleuser']
# c.Spawner.cmd = ['jupyterhub-singleuser']

## Maximum number of consecutive failures to allow before shutting down
#  JupyterHub.
#
#  This helps JupyterHub recover from a certain class of problem preventing
#  launch in contexts where the Hub is automatically restarted (e.g. systemd,
#  docker, kubernetes).
#
#  A limit of 0 means no limit and consecutive failures will not be tracked.
#  Default: 0
# c.Spawner.consecutive_failure_limit = 0

## Minimum number of cpu-cores a single-user notebook server is guaranteed to
#  have available.
#
#  If this value is set to 0.5, allows use of 50% of one CPU. If this value is
#  set to 2, allows use of up to 2 CPUs.
#
#  **This is a configuration setting. Your spawner must implement support for the
#  limit to work.** The default spawner, `LocalProcessSpawner`, does **not**
#  implement this support. A custom spawner **must** add support for this setting
#  for it to be enforced.
#  Default: None
# c.Spawner.cpu_guarantee = None

## Maximum number of cpu-cores a single-user notebook server is allowed to use.
#
#  If this value is set to 0.5, allows use of 50% of one CPU. If this value is
#  set to 2, allows use of up to 2 CPUs.
#
#  The single-user notebook server will never be scheduled by the kernel to use
#  more cpu-cores than this. There is no guarantee that it can access this many
#  cpu-cores.
#
#  **This is a configuration setting. Your spawner must implement support for the
#  limit to work.** The default spawner, `LocalProcessSpawner`, does **not**
#  implement this support. A custom spawner **must** add support for this setting
#  for it to be enforced.
#  Default: None
# c.Spawner.cpu_limit = None

## Enable debug-logging of the single-user server
#  Default: False
# c.Spawner.debug = False

## The URL the single-user server should start in.
#
#  `{username}` will be expanded to the user's username
#
#  Example uses:
#
#  - You can set `notebook_dir` to `/` and `default_url` to `/tree/home/{username}` to allow people to
#    navigate the whole filesystem from their notebook server, but still start in their home directory.
#  - Start with `/notebooks` instead of `/tree` if `default_url` points to a notebook instead of a directory.
#  - You can set this to `/lab` to have JupyterLab start by default, rather than Jupyter Notebook.
#  Default: ''
# c.Spawner.default_url = ''

## Disable per-user configuration of single-user servers.
#
#  When starting the user's single-user server, any config file found in the
#  user's $HOME directory will be ignored.
#
#  Note: a user could circumvent this if the user modifies their Python
#  environment, such as when they have their own conda environments / virtualenvs
#  / containers.
#  Default: False
# c.Spawner.disable_user_config = False

## List of environment variables for the single-user server to inherit from the
#  JupyterHub process.
#
#  This list is used to ensure that sensitive information in the JupyterHub
#  process's environment (such as `CONFIGPROXY_AUTH_TOKEN`) is not passed to the
#  single-user server's process.
#  Default: ['PATH', 'PYTHONPATH', 'CONDA_ROOT', 'CONDA_DEFAULT_ENV', 'VIRTUAL_ENV', 'LANG', 'LC_ALL', 'JUPYTERHUB_SINGLEUSER_APP']
# c.Spawner.env_keep = ['PATH', 'PYTHONPATH', 'CONDA_ROOT', 'CONDA_DEFAULT_ENV', 'VIRTUAL_ENV', 'LANG', 'LC_ALL', 'JUPYTERHUB_SINGLEUSER_APP']

## Extra environment variables to set for the single-user server's process.
#
#  Environment variables that end up in the single-user server's process come from 3 sources:
#    - This `environment` configurable
#    - The JupyterHub process' environment variables that are listed in `env_keep`
#    - Variables to establish contact between the single-user notebook and the hub (such as JUPYTERHUB_API_TOKEN)
#
#  The `environment` configurable should be set by JupyterHub administrators to
#  add installation specific environment variables. It is a dict where the key is
#  the name of the environment variable, and the value can be a string or a
#  callable. If it is a callable, it will be called with one parameter (the
#  spawner instance), and should return a string fairly quickly (no blocking
#  operations please!).
#
#  Note that the spawner class' interface is not guaranteed to be exactly same
#  across upgrades, so if you are using the callable take care to verify it
#  continues to work after upgrades!
#
#  .. versionchanged:: 1.2
#      environment from this configuration has highest priority,
#      allowing override of 'default' env variables,
#      such as JUPYTERHUB_API_URL.
#  Default: {}
# c.Spawner.environment = {}

## Timeout (in seconds) before giving up on a spawned HTTP server
#
#  Once a server has successfully been spawned, this is the amount of time we
#  wait before assuming that the server is unable to accept connections.
#  Default: 30
# c.Spawner.http_timeout = 30

## The IP address (or hostname) the single-user server should listen on.
#
#  The JupyterHub proxy implementation should be able to send packets to this
#  interface.
#  Default: ''
# c.Spawner.ip = ''

## Minimum number of bytes a single-user notebook server is guaranteed to have
#  available.
#
#  Allows the following suffixes:
#    - K -> Kilobytes
#    - M -> Megabytes
#    - G -> Gigabytes
#    - T -> Terabytes
#
#  **This is a configuration setting. Your spawner must implement support for the
#  limit to work.** The default spawner, `LocalProcessSpawner`, does **not**
#  implement this support. A custom spawner **must** add support for this setting
#  for it to be enforced.
#  Default: None
# c.Spawner.mem_guarantee = None

## Maximum number of bytes a single-user notebook server is allowed to use.
#
#  Allows the following suffixes:
#    - K -> Kilobytes
#    - M -> Megabytes
#    - G -> Gigabytes
#    - T -> Terabytes
#
#  If the single user server tries to allocate more memory than this, it will
#  fail. There is no guarantee that the single-user notebook server will be able
#  to allocate this much memory - only that it can not allocate more than this.
#
#  **This is a configuration setting. Your spawner must implement support for the
#  limit to work.** The default spawner, `LocalProcessSpawner`, does **not**
#  implement this support. A custom spawner **must** add support for this setting
#  for it to be enforced.
#  Default: None
# c.Spawner.mem_limit = None

## Path to the notebook directory for the single-user server.
#
#  The user sees a file listing of this directory when the notebook interface is
#  started. The current interface does not easily allow browsing beyond the
#  subdirectories in this directory's tree.
#
#  `~` will be expanded to the home directory of the user, and {username} will be
#  replaced with the name of the user.
#
#  Note that this does *not* prevent users from accessing files outside of this
#  path! They can do so with many other means.
#  Default: ''
# c.Spawner.notebook_dir = ''

## An HTML form for options a user can specify on launching their server.
#
#  The surrounding `<form>` element and the submit button are already provided.
#
#  For example:
#
#  .. code:: html
#
#      Set your key:
#      <input name="key" val="default_key"></input>
#      <br>
#      Choose a letter:
#      <select name="letter" multiple="true">
#        <option value="A">The letter A</option>
#        <option value="B">The letter B</option>
#      </select>
#
#  The data from this form submission will be passed on to your spawner in
#  `self.user_options`
#
#  Instead of a form snippet string, this could also be a callable that takes as
#  one parameter the current spawner instance and returns a string. The callable
#  will be called asynchronously if it returns a future, rather than a str. Note
#  that the interface of the spawner class is not deemed stable across versions,
#  so using this functionality might cause your JupyterHub upgrades to break.
#  Default: traitlets.Undefined
# c.Spawner.options_form = traitlets.Undefined

## Interpret HTTP form data
#
#  Form data will always arrive as a dict of lists of strings. Override this
#  function to understand single-values, numbers, etc.
#
#  This should coerce form data into the structure expected by self.user_options,
#  which must be a dict, and should be JSON-serializeable, though it can contain
#  bytes in addition to standard JSON data types.
#
#  This method should not have any side effects. Any handling of `user_options`
#  should be done in `.start()` to ensure consistent behavior across servers
#  spawned via the API and form submission page.
#
#  Instances will receive this data on self.user_options, after passing through
#  this function, prior to `Spawner.start`.
#
#  .. versionchanged:: 1.0
#      user_options are persisted in the JupyterHub database to be reused
#      on subsequent spawns if no options are given.
#      user_options is serialized to JSON as part of this persistence
#      (with additional support for bytes in case of uploaded file data),
#      and any non-bytes non-jsonable values will be replaced with None
#      if the user_options are re-used.
#  Default: traitlets.Undefined
# c.Spawner.options_from_form = traitlets.Undefined

## Interval (in seconds) on which to poll the spawner for single-user server's
#  status.
#
#  At every poll interval, each spawner's `.poll` method is called, which checks
#  if the single-user server is still running. If it isn't running, then
#  JupyterHub modifies its own state accordingly and removes appropriate routes
#  from the configurable proxy.
#  Default: 30
# c.Spawner.poll_interval = 30

## The port for single-user servers to listen on.
#
#  Defaults to `0`, which uses a randomly allocated port number each time.
#
#  If set to a non-zero value, all Spawners will use the same port, which only
#  makes sense if each server is on a different address, e.g. in containers.
#
#  New in version 0.7.
#  Default: 0
# c.Spawner.port = 0

## An optional hook function that you can implement to do work after the spawner
#  stops.
#
#  This can be set independent of any concrete spawner implementation.
#  Default: None
# c.Spawner.post_stop_hook = None

## An optional hook function that you can implement to do some bootstrapping work
#  before the spawner starts. For example, create a directory for your user or
#  load initial content.
#
#  This can be set independent of any concrete spawner implementation.
#
#  This maybe a coroutine.
#
#  Example::
#
#      from subprocess import check_call
#      def my_hook(spawner):
#          username = spawner.user.name
#          check_call(['./examples/bootstrap-script/bootstrap.sh', username])
#
#      c.Spawner.pre_spawn_hook = my_hook
#  Default: None
# c.Spawner.pre_spawn_hook = None

## List of SSL alt names
#
#  May be set in config if all spawners should have the same value(s), or set at
#  runtime by Spawner that know their names.
#  Default: []
# c.Spawner.ssl_alt_names = []

## Whether to include DNS:localhost, IP:127.0.0.1 in alt names
#  Default: True
# c.Spawner.ssl_alt_names_include_local = True

## Timeout (in seconds) before giving up on starting of single-user server.
#
#  This is the timeout for start to return, not the timeout for the server to
#  respond. Callers of spawner.start will assume that startup has failed if it
#  takes longer than this. start should return when the server process is started
#  and its location is known.
#  Default: 60
# c.Spawner.start_timeout = 60

# ------------------------------------------------------------------------------
# Authenticator(LoggingConfigurable) configuration
# ------------------------------------------------------------------------------
## Base class for implementing an authentication provider for JupyterHub

## Set of users that will have admin rights on this JupyterHub.
#
#  Admin users have extra privileges:
#   - Use the admin panel to see list of users logged in
#   - Add / remove users in some authenticators
#   - Restart / halt the hub
#   - Start / stop users' single-user servers
#   - Can access each individual users' single-user server (if configured)
#
#  Admin access should be treated the same way root access is.
#
#  Defaults to an empty set, in which case no user has admin access.
#  Default: set()
# c.Authenticator.admin_users = set()

## Set of usernames that are allowed to log in.
#
#  Use this with supported authenticators to restrict which users can log in.
#  This is an additional list that further restricts users, beyond whatever
#  restrictions the authenticator has in place.
#
#  If empty, does not perform any additional restriction.
#
#  .. versionchanged:: 1.2
#      `Authenticator.whitelist` renamed to `allowed_users`
#  Default: set()
# c.Authenticator.allowed_users = set()

## The max age (in seconds) of authentication info before forcing a refresh of
#  user auth info.
#
#  Refreshing auth info allows, e.g. requesting/re-validating auth tokens.
#
#  See :meth:`.refresh_user` for what happens when user auth info is refreshed
#  (nothing by default).
#  Default: 300
# c.Authenticator.auth_refresh_age = 300

## Automatically begin the login process
#
#  rather than starting with a "Login with..." link at `/hub/login`
#
#  To work, `.login_url()` must give a URL other than the default `/hub/login`,
#  such as an oauth handler or another automatic login handler, registered with
#  `.get_handlers()`.
#
#  .. versionadded:: 0.8
#  Default: False
# c.Authenticator.auto_login = False

## Set of usernames that are not allowed to log in.
#
#  Use this with supported authenticators to restrict which users can not log in.
#  This is an additional block list that further restricts users, beyond whatever
#  restrictions the authenticator has in place.
#
#  If empty, does not perform any additional restriction.
#
#  .. versionadded: 0.9
#
#  .. versionchanged:: 1.2
#      `Authenticator.blacklist` renamed to `blocked_users`
#  Default: set()
# c.Authenticator.blocked_users = set()

## Delete any users from the database that do not pass validation
#
#  When JupyterHub starts, `.add_user` will be called on each user in the
#  database to verify that all users are still valid.
#
#  If `delete_invalid_users` is True, any users that do not pass validation will
#  be deleted from the database. Use this if users might be deleted from an
#  external system, such as local user accounts.
#
#  If False (default), invalid users remain in the Hub's database and a warning
#  will be issued. This is the default to avoid data loss due to config changes.
#  Default: False
# c.Authenticator.delete_invalid_users = False

## Enable persisting auth_state (if available).
#
#  auth_state will be encrypted and stored in the Hub's database. This can
#  include things like authentication tokens, etc. to be passed to Spawners as
#  environment variables.
#
#  Encrypting auth_state requires the cryptography package.
#
#  Additionally, the JUPYTERHUB_CRYPT_KEY environment variable must contain one
#  (or more, separated by ;) 32B encryption keys. These can be either base64 or
#  hex-encoded.
#
#  If encryption is unavailable, auth_state cannot be persisted.
#
#  New in JupyterHub 0.8
#  Default: False
# c.Authenticator.enable_auth_state = False

## An optional hook function that you can implement to do some bootstrapping work
#  during authentication. For example, loading user account details from an
#  external system.
#
#  This function is called after the user has passed all authentication checks
#  and is ready to successfully authenticate. This function must return the
#  authentication dict reguardless of changes to it.
#
#  This maybe a coroutine.
#
#  .. versionadded: 1.0
#
#  Example::
#
#      import os, pwd
#      def my_hook(authenticator, handler, authentication):
#          user_data = pwd.getpwnam(authentication['name'])
#          spawn_data = {
#              'pw_data': user_data
#              'gid_list': os.getgrouplist(authentication['name'], user_data.pw_gid)
#          }
#
#          if authentication['auth_state'] is None:
#              authentication['auth_state'] = {}
#          authentication['auth_state']['spawn_data'] = spawn_data
#
#          return authentication
#
#      c.Authenticator.post_auth_hook = my_hook
#  Default: None
# c.Authenticator.post_auth_hook = None

## Force refresh of auth prior to spawn.
#
#  This forces :meth:`.refresh_user` to be called prior to launching a server, to
#  ensure that auth state is up-to-date.
#
#  This can be important when e.g. auth tokens that may have expired are passed
#  to the spawner via environment variables from auth_state.
#
#  If refresh_user cannot refresh the user auth data, launch will fail until the
#  user logs in again.
#  Default: False
# c.Authenticator.refresh_pre_spawn = False

## Dictionary mapping authenticator usernames to JupyterHub users.
#
#  Primarily used to normalize OAuth user names to local users.
#  Default: {}
# c.Authenticator.username_map = {}

## Regular expression pattern that all valid usernames must match.
#
#  If a username does not match the pattern specified here, authentication will
#  not be attempted.
#
#  If not set, allow any username.
#  Default: ''
# c.Authenticator.username_pattern = ''

## Deprecated, use `Authenticator.allowed_users`
#  Default: set()
# c.Authenticator.whitelist = set()

# ------------------------------------------------------------------------------
# CryptKeeper(SingletonConfigurable) configuration
# ------------------------------------------------------------------------------
## Encapsulate encryption configuration
#
#  Use via the encryption_config singleton below.

#  Default: []
# c.CryptKeeper.keys = []

## The number of threads to allocate for encryption
#  Default: 8
# c.CryptKeeper.n_threads = 8

# ------------------------------------------------------------------------------
# Pagination(Configurable) configuration
# ------------------------------------------------------------------------------
## Default number of entries per page for paginated results.
#  Default: 100
# c.Pagination.default_per_page = 100

## Maximum number of entries per page for paginated results.
#  Default: 250
# c.Pagination.max_per_page = 250



# Copyright 2020 QuantStack
# Distributed under the terms of the Modified BSD License.

from __future__ import annotations

import json
import uuid
from datetime import date, datetime
from enum import Enum
from typing import Generic, List, Optional, TypeVar

from pydantic import BaseModel, Field, root_validator, validator
from pydantic.generics import GenericModel

T = TypeVar('T')


class BaseProfile(BaseModel):
    name: Optional[str] = Field(None, nullable=True)
    avatar_url: str

    class Config:
        orm_mode = True


class Profile(BaseProfile):
    user: BaseUser


class BaseUser(BaseModel):
    id: uuid.UUID
    username: str

    class Config:
        orm_mode = True


class User(BaseUser):
    profile: BaseProfile


Profile.update_forward_refs()


Role = Field(None, regex='owner|maintainer|member')


class Member(BaseModel):
    role: str = Role
    user: User

    class Config:
        orm_mode = True


class Pagination(BaseModel):
    skip: int = Field(0, title='The number of skipped records')
    limit: int = Field(0, title='The maximum number of returned records')
    all_records_count: int = Field(0, title="The number of available records")


class MirrorMode(str, Enum):
    proxy = "proxy"
    mirror = "mirror"


class ChannelBase(BaseModel):
    name: str = Field(None, title='The name of the channel', max_length=50)
    description: Optional[str] = Field(
        None, title='The description of the channel', max_length=300, nullable=True
    )
    private: bool = Field(True, title="channel should be private")
    size_limit: Optional[int] = Field(
        None, title="size limit of the channel", nullable=True
    )
    ttl: int = Field(36000, title="ttl of the channel")
    mirror_channel_url: Optional[str] = Field(
        None, regex="^(http|https)://.+", nullable=True
    )
    mirror_mode: Optional[MirrorMode] = Field(None, nullable=True)

    @validator("size_limit")
    def check_positive(cls, v):
        if v is not None and v < 0:
            return ValueError("must be positive value")
        return v

    class Config:
        orm_mode = True


class ChannelExtra(ChannelBase):
    members_count: int
    packages_count: int


class ChannelRole(BaseModel):
    name: str = Field(title="channel name")
    role: str = Field(title="user role")

    class Config:
        orm_mode = True


class ChannelActionEnum(str, Enum):
    """Execute special actions on channels (they may need specific permissions):

    * `synchronize` -- _mirror only_, synchronize mirror channels and extract metadata
      from packages (compute heavy)
    * `synchronize_repodata` -- _mirror only_, synchronize mirror channels by extracting
      metadata from index (less heavy)
    * `reindex` -- find all downloaded packages and re-create db and index
    * `generate_indexes` -- generate indexes (repodata.json and other) from data in db
    * `validate_packages` -- validate package files
    * `synchronize_metrics` -- _non-mirror_, pull download metrics from known mirrors
    * `cleanup` -- fix inconsistencies in database and pkgstore
    * `cleanup_dry_run` -- display what changes `cleanup` would do
    """

    synchronize = 'synchronize'
    synchronize_repodata = "synchronize_repodata"
    reindex = 'reindex'
    generate_indexes = 'generate_indexes'
    validate_packages = 'validate_packages'
    synchronize_metrics = 'synchronize_metrics'
    cleanup = 'cleanup'
    cleanup_dry_run = 'cleanup_dry_run'

    # handlers for new actions should be registered in quetz.job.handlers


class ChannelMetadata(BaseModel):
    includelist: Optional[List[str]] = Field(
        None,
        title="list of packages to include while creating a channel",
        nullable=True,
    )
    excludelist: Optional[List[str]] = Field(
        None,
        title="list of packages to exclude while creating a channel",
        nullable=True,
    )
    proxylist: Optional[List[str]] = Field(
        None,
        title="list of packages that should only be proxied (not copied, "
        "stored and redistributed)",
        nullable=True,
    )


class Channel(ChannelBase):
    metadata: ChannelMetadata = Field(
        default_factory=ChannelMetadata, title="channel metadata", example={}
    )

    actions: Optional[List[ChannelActionEnum]] = Field(
        None,
        title="list of actions to run after channel creation "
        "(see /channels/{}/actions for description)",
        nullable=True,
    )

    @root_validator
    def check_mirror_params(cls, values):
        mirror_url = values.get("mirror_channel_url")
        mirror_mode = values.get("mirror_mode")

        if mirror_url and not mirror_mode:
            raise ValueError(
                "'mirror_channel_url' provided but 'mirror_mode' is undefined"
            )
        if not mirror_url and mirror_mode:
            raise ValueError(
                "'mirror_mode' provided but 'mirror_channel_url' is undefined"
            )

        return values


class ChannelMirrorBase(BaseModel):
    url: str = Field(None, regex="^(http|https)://.+")
    api_endpoint: Optional[str] = Field(None, regex="^(http|https)://.+", nullable=True)
    metrics_endpoint: Optional[str] = Field(
        None, regex="^(http|https)://.+", nullable=True
    )

    class Config:
        orm_mode = True


class ChannelMirror(ChannelMirrorBase):
    id: uuid.UUID


class Package(BaseModel):
    name: str = Field(
        None, title='The name of package', max_length=1500, regex=r'^[a-z0-9-_\.]*$'
    )
    summary: Optional[str] = Field(
        None, title='The summary of the package', nullable=True
    )
    description: Optional[str] = Field(
        None, title='The description of the package', nullable=True
    )
    url: Optional[str] = Field(None, title="project url", nullable=True)
    platforms: List[str] = Field(None, title="supported platforms", nullable=True)
    current_version: Optional[str] = Field(
        None, title="latest version of any platform", nullable=True
    )
    latest_change: Optional[datetime] = Field(
        None, title="date of latest change", nullable=True
    )

    @validator("platforms", pre=True)
    def parse_list_of_platforms(cls, v):
        if isinstance(v, str):
            return v.split(":")
        else:
            return v

    class Config:
        orm_mode = True


class PackageRole(BaseModel):
    name: str = Field(title='The name of package')
    channel_name: str = Field(title='The channel this package belongs to')
    role: str = Field(title="user role for this package")

    class Config:
        orm_mode = True


class PackageSearch(Package):
    channel_name: str = Field(None, title='The channel this package belongs to')


class ChannelSearch(BaseModel):
    name: str = Field(None, title='The name of the channel', max_length=1500)
    description: str = Field(None, title='The description of the channel')
    private: bool = Field(None, title='The visibility of the channel')

    class Config:
        orm_mode = True


class PaginatedResponse(GenericModel, Generic[T]):
    pagination: Pagination = Field(None, title="Pagination object")
    result: List[T] = Field([], title="Result objects")


class PostMember(BaseModel):
    username: str
    role: str = Role


class UserRole(BaseModel):
    role: str = Role


class CPRole(BaseModel):
    channel: str
    package: Optional[str] = Field(None, nullable=True)
    role: str = Role


class BaseApiKey(BaseModel):
    description: str
    time_created: Optional[date] = Field(None, nullable=True)
    expire_at: Optional[date] = Field(None, nullable=True)
    roles: Optional[List[CPRole]] = Field(None, nullable=True)


class ApiKey(BaseApiKey):
    key: str


class PackageVersion(BaseModel):
    id: uuid.UUID
    channel_name: str
    package_name: str
    platform: str
    version: str
    build_string: str
    build_number: int

    filename: str
    info: dict
    uploader: BaseProfile
    time_created: datetime
    download_count: int

    class Config:
        orm_mode = True

    @validator("uploader", pre=True)
    def convert_uploader(cls, v):
        if hasattr(v, "profile"):
            return v.profile
        else:
            return v

    @validator("info", pre=True)
    def load_json(cls, v):
        if isinstance(v, str):
            return json.loads(v)
        else:
            return v


class ChannelAction(BaseModel):
    action: ChannelActionEnum
    start_at: Optional[datetime] = Field(None, nullable=True)
    repeat_every_seconds: Optional[int] = Field(None, nullable=True)



"""configure mirror endpoints

Revision ID: 303ff70c27fc
Revises: 8dfb7c4bfbd7
Create Date: 2020-12-22 17:35:45.164946

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = '303ff70c27fc'
down_revision = '8dfb7c4bfbd7'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        'channel_mirrors', sa.Column('api_endpoint', sa.String(), nullable=True)
    )
    op.add_column(
        'channel_mirrors', sa.Column('metrics_endpoint', sa.String(), nullable=True)
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('channel_mirrors', 'metrics_endpoint')
    op.drop_column('channel_mirrors', 'api_endpoint')
    # ### end Alembic commands ###



import logging
import os
from pathlib import Path

from sqlalchemy import event
from sqlalchemy.ext.compiler import compiles
from sqlalchemy.pool import Pool, StaticPool
from sqlalchemy.sql import expression
from sqlalchemy.types import Numeric

logger = logging.getLogger("quetz")

try:
    from sqlite3 import Connection as SQLiteConnection

    sqlite_available = True
except ImportError:
    sqlite_available = False
    print("Could not import sqlite3")


try:
    from psycopg2.extensions import connection as PGConnection

    postgres_available = True
except ImportError:
    postgres_available = False

    print("Could not import postgres backend")


pg_create_function = """\
CREATE OR REPLACE FUNCTION version_compare(varchar, varchar)
RETURNS boolean
AS '{libpath}'
LANGUAGE 'c'
"""

search_path = os.getenv("QUETZ_DATABASE_PLUGIN_PATH")

sqlite_plugin, pg_plugin = None, None
if search_path:
    plugin_path = Path(search_path)

    if (plugin_path / "libquetz_pg.so").exists():
        pg_plugin = str((plugin_path / "libquetz_pg.so").resolve())
        logger.info(
            "Looking for database extension: "
            f"{plugin_path / 'libquetz_pg.so'}: FOUND"
        )
    else:
        logger.info(
            "Looking for database extension: "
            f"{plugin_path / 'libquetz_pg.so'}: NOT FOUND"
        )

    if (plugin_path / "libquetz_sqlite.so").exists():
        sqlite_plugin = str((plugin_path / "libquetz_sqlite.so").resolve())
        logger.info(
            "Looking for database extension: "
            f"{plugin_path / 'libquetz_sqlite.so'}: FOUND"
        )
    else:
        logger.info(
            "Looking for database extension: "
            f"{plugin_path / 'libquetz_sqlite.so'}: NOT FOUND"
        )


class _version_match(expression.FunctionElement):
    type = Numeric()
    name = 'version_match'


@compiles(_version_match, 'sqlite')
def sqlite_version_match(element, compiler, **kw):
    return compiler.visit_function(element)


if not sqlite_plugin and not pg_plugin:
    version_match = None
else:
    version_match = _version_match


def load_plugins_after_connect(dbapi_connection, connection_record):
    if sqlite_available and type(dbapi_connection) is SQLiteConnection:
        if sqlite_plugin:
            dbapi_connection.enable_load_extension(True)
            dbapi_connection.load_extension(sqlite_plugin)
            dbapi_connection.enable_load_extension(False)
    elif postgres_available and type(dbapi_connection) is PGConnection:
        if pg_plugin:
            cursor = dbapi_connection.cursor()
            cursor.execute(pg_create_function.format(libpath=pg_plugin))


event.listen(Pool, 'first_connect', load_plugins_after_connect)
event.listen(StaticPool, 'first_connect', load_plugins_after_connect)



import uuid
from pathlib import Path

import pytest

from quetz.db_models import Identity, Profile, User
from quetz.rest_models import Channel, Package


@pytest.fixture
def private_channel(dao, other_user):
    channel_name = "private-channel"

    channel_data = Channel(name=channel_name, private=True)
    channel = dao.create_channel(channel_data, other_user.id, "owner")

    return channel


@pytest.fixture
def private_package(dao, other_user, private_channel):
    package_name = "private-package"
    package_data = Package(name=package_name)
    package = dao.create_package(
        private_channel.name, package_data, other_user.id, "owner"
    )

    return package


@pytest.fixture
def private_package_version(
    dao, private_channel, private_package, other_user, config, package_name
):
    package_format = "tarbz2"
    package_info = "{}"
    channel_name = private_channel.name
    filename = Path("test-package-0.1-0.tar.bz2")

    pkgstore = config.get_package_store()
    with open(filename, "rb") as fid:
        pkgstore.add_file(fid.read(), channel_name, "linux-64" / filename)

    platform = "linux-64"
    version = dao.create_version(
        private_channel.name,
        private_package.name,
        package_format,
        platform,
        "0.1",
        "0",
        "",
        str(filename),
        package_info,
        other_user.id,
        size=0,
    )

    dao.update_package_channeldata(
        private_channel.name,
        private_package.name,
        {'name': package_name, 'subdirs': [platform]},
    )

    return version


@pytest.fixture()
def other_user_without_profile(db):
    user = User(id=uuid.uuid4().bytes, username="other")
    db.add(user)
    return user


@pytest.fixture
def other_user(other_user_without_profile, db):
    profile = Profile(
        name="Other", avatar_url="http:///avatar", user=other_user_without_profile
    )
    identity = Identity(
        provider="github",
        identity_id="github",
        username="btel",
        user=other_user_without_profile,
    )
    db.add(profile)
    db.add(identity)
    db.commit()
    yield other_user_without_profile


@pytest.fixture
def pkgstore(config):
    return config.get_package_store()



"""register mirrors

Revision ID: 98c04a65df4a
Revises: 8d1e9a9e0b1f
Create Date: 2020-12-21 11:10:15.107635

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = '98c04a65df4a'
down_revision = '8d1e9a9e0b1f'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        'channel_mirrors',
        sa.Column('id', sa.LargeBinary(length=16), nullable=False),
        sa.Column('channel_name', sa.String(), nullable=True),
        sa.Column('url', sa.String(), nullable=False),
        sa.Column('last_synchronised', sa.DateTime(), nullable=True),
        sa.ForeignKeyConstraint(
            ['channel_name'],
            ['channels.name'],
        ),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('url'),
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('channel_mirrors')
    # ### end Alembic commands ###



# Copyright 2020 Codethink Ltd
# Distributed under the terms of the Modified BSD License.

import json

from quetz import db_models


def export(dao, channel_name, subdir):
    repodata = {
        "info": {"subdir": subdir},
        "packages": {},
        "packages.conda": {},
        "repodata_version": 1,
    }
    if dao.is_active_platform(channel_name, subdir):
        packages = repodata["packages"]
        packages_conda = repodata["packages.conda"]

        for filename, info, format, time_modified in dao.get_package_infos(
            channel_name, subdir
        ):
            data = json.loads(info)
            data['time_modified'] = int(time_modified.timestamp())
            if format == db_models.PackageFormatEnum.conda:
                packages_conda[filename] = data
            else:
                packages[filename] = data

        return repodata
    else:
        return repodata



"""
Define common dependencies for fastapi dependency injection system
"""

import logging

import requests
from fastapi import BackgroundTasks, Depends, HTTPException, Request, status
from requests.adapters import HTTPAdapter
from sqlalchemy.orm import Session
from urllib3.util.retry import Retry

from quetz import authorization, db_models
from quetz.config import Config
from quetz.dao import Dao
from quetz.database import get_session as get_db_session
from quetz.tasks.common import Task

DEFAULT_TIMEOUT = 5  # seconds
MAX_RETRIES = 3

logger = logging.getLogger("quetz")


class TimeoutHTTPAdapter(HTTPAdapter):
    def __init__(self, *args, **kwargs):
        self.timeout = DEFAULT_TIMEOUT
        if "timeout" in kwargs:
            self.timeout = kwargs["timeout"]
            del kwargs["timeout"]
        super().__init__(*args, **kwargs)

    def send(self, request, **kwargs):
        timeout = kwargs.get("timeout")
        if timeout is None:
            kwargs["timeout"] = self.timeout
        return super().send(request, **kwargs)


def get_config():
    config = Config()
    return config


def get_db(config: Config = Depends(get_config)):
    database_url = config.sqlalchemy_database_url
    db = get_db_session(database_url, echo=config.sqlalchemy_echo_sql)
    try:
        yield db
    finally:
        db.close()


def get_dao(db: Session = Depends(get_db)):
    return Dao(db)


def get_session(request: Request):
    return request.session


def get_remote_session() -> requests.Session:
    session = requests.Session()
    retries = Retry(
        total=MAX_RETRIES, backoff_factor=1, status_forcelist=[429, 500, 502, 503, 504]
    )

    adapter = TimeoutHTTPAdapter(max_retries=retries)

    session.mount("https://", adapter)
    session.mount("http://", adapter)

    return session


def get_rules(
    request: Request,
    session: dict = Depends(get_session),
    db: Session = Depends(get_db),
):
    return authorization.Rules(request.headers.get("x-api-key"), session, db)


def get_tasks_worker(
    background_tasks: BackgroundTasks,
    dao: Dao = Depends(get_dao),
    auth: authorization.Rules = Depends(get_rules),
    session: requests.Session = Depends(get_remote_session),
    config: Config = Depends(get_config),
) -> Task:
    return Task(auth, dao.db)


class ChannelChecker:
    def __init__(
        self,
        allow_proxy: bool = False,
        allow_mirror: bool = False,
        allow_local: bool = True,
    ):
        self.allow_proxy = allow_proxy
        self.allow_mirror = allow_mirror
        self.allow_local = allow_local

    def __call__(
        self,
        channel_name: str,
        dao: Dao = Depends(get_dao),
        auth: authorization.Rules = Depends(get_rules),
    ) -> db_models.Channel:
        channel = dao.get_channel(channel_name.lower())

        if not channel:
            raise HTTPException(
                status_code=status.HTTP_404_NOT_FOUND,
                detail=f"Channel {channel_name} not found",
            )

        auth.assert_channel_read(channel)

        mirror_url = channel.mirror_channel_url

        is_proxy = mirror_url and channel.mirror_mode == "proxy"
        is_mirror = mirror_url and channel.mirror_mode == "mirror"
        is_local = not mirror_url
        if is_proxy and not self.allow_proxy:
            raise HTTPException(
                status_code=status.HTTP_405_METHOD_NOT_ALLOWED,
                detail="This method is not implemented for proxy channels",
            )
        if is_mirror and not self.allow_mirror:
            raise HTTPException(
                status_code=status.HTTP_405_METHOD_NOT_ALLOWED,
                detail="This method is not implemented for mirror channels",
            )
        if is_local and not self.allow_local:
            raise HTTPException(
                status_code=status.HTTP_405_METHOD_NOT_ALLOWED,
                detail="This method is not implemented for local channels",
            )
        return channel


get_channel_or_fail = ChannelChecker(allow_proxy=False, allow_mirror=True)
get_channel_allow_proxy = ChannelChecker(allow_proxy=True, allow_mirror=True)
get_channel_mirror_only = ChannelChecker(allow_mirror=True, allow_local=False)


def get_package_or_fail(
    package_name: str,
    channel_name: str,
    dao: Dao = Depends(get_dao),
    auth: authorization.Rules = Depends(get_rules),
) -> db_models.Package:
    package = dao.get_package(channel_name.lower(), package_name)

    if not package:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=f"Package {channel_name}/{package_name} not found",
        )

    auth.assert_package_read(package)
    return package



from .oauth2 import OAuthAuthenticator


class GithubAuthenticator(OAuthAuthenticator):
    """Use Github account to authenticate users with Quetz.

    To enable add the following to the configuration file:

    .. code::

      [github]
      client_id = "fde330aef1fbe39991"
      client_secret = "03728444a12abff17e9444fd231b4379d58f0b"

    You can obtain ``client_id`` and ``client_secret`` by registering your
    application with Github at this URL:
    `<https://github.com/settings/applications/new>`_.
    """

    provider = "github"
    collect_emails = False

    # oauth client params
    access_token_url = 'https://github.com/login/oauth/access_token'
    authorize_url = 'https://github.com/login/oauth/authorize'
    api_base_url = 'https://api.github.com/'
    scope = 'user:email'

    # endpoint urls
    validate_token_url = "user"
    revoke_url = 'https://github.com/settings/connections/applications/{client_id}'

    async def userinfo(self, request, token):
        resp = await self.client.get('user', token=token)
        profile = resp.json()

        if self.collect_emails:
            emails = await self.client.get('user/emails', token=token)
            profile["emails"] = emails.json()

        return profile

    def configure(self, config):
        if config.configured_section("github"):
            self.client_id = config.github_client_id
            self.client_secret = config.github_client_secret
            self.is_enabled = True
            if config.configured_section("users"):
                self.collect_emails = config.users_collect_emails

        else:
            self.is_enabled = False

        # call the configure of base class to set default_channel and default role
        super().configure(config)



import os
import socket
from contextlib import closing

import pytest
import requests
from fastapi import BackgroundTasks

from quetz.authorization import Rules
from quetz.dao import Dao
from quetz.db_models import User
from quetz.tasks.workers import RQManager, SubprocessWorker, ThreadingWorker


@pytest.fixture
def sqlite_url(config_dir):
    # overriding sqlite_url to save to file so that
    # we can access the same db from a sub-process
    return f'sqlite:///{config_dir}/quetz.db'


@pytest.fixture
def http_session():
    return requests.Session()


@pytest.fixture
def background_tasks():
    bg_tasks = BackgroundTasks()
    return bg_tasks


@pytest.fixture
def api_key():
    return "api-key"


@pytest.fixture
def browser_session():
    return {}


@pytest.fixture
def auth(db, api_key, browser_session):
    return Rules(api_key, browser_session, db)


@pytest.fixture
def redis_ip():
    return "127.0.0.1"


@pytest.fixture
def redis_port():
    return 6379


@pytest.fixture
def redis_db():
    return 0


@pytest.fixture
def threading_worker(config):
    worker = ThreadingWorker(config)
    return worker


@pytest.fixture
def subprocess_worker(api_key, browser_session, db, config):
    SubprocessWorker._executor = None
    worker = SubprocessWorker(config)
    return worker


@pytest.fixture
def redis_worker(redis_ip, redis_port, redis_db, api_key, browser_session, db, config):
    worker = RQManager(
        redis_ip,
        redis_port,
        redis_db,
        config,
        no_testing=False,
    )
    return worker


def check_socket(host, port):
    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as sock:
        sock.settimeout(2)
        if sock.connect_ex((host, port)) == 0:
            return True
        else:
            return False


def check_redis():
    return check_socket('127.0.0.1', 6379)


@pytest.fixture(
    params=[
        "threading_worker",
        "subprocess_worker",
        pytest.param(  # type: ignore
            "redis_worker",
            marks=pytest.mark.skipif(not check_redis(), reason='no redis'),
        ),
    ]
)
def any_worker(request):
    val = request.getfixturevalue(request.param)
    return val


def basic_function(config_dir):
    os.chdir(config_dir)
    with open("test.txt", "w") as fid:
        fid.write("hello world!")


def function_with_dao(dao: Dao):
    dao.create_user_with_role("my-user")


@pytest.fixture
def db_cleanup(config):
    # we can't use the db fixture for cleaning up because
    # it automatically rollsback all operations

    yield

    from quetz.database import get_session

    db = get_session(config.sqlalchemy_database_url)
    user = db.query(User).one_or_none()
    if user:
        db.delete(user)
        db.commit()


@pytest.mark.asyncio
async def test_threading_worker_execute(background_tasks, any_worker, db, config_dir):
    any_worker.execute(basic_function, config_dir=config_dir)

    await any_worker.wait()

    with open("test.txt") as fid:
        output = fid.read()

    assert output == "hello world!"


@pytest.mark.asyncio
async def test_threading_worker_execute_with_dao(
    background_tasks, any_worker, db, db_cleanup
):
    any_worker.execute(function_with_dao)

    await any_worker.wait()

    users = db.query(User).all()

    assert len(users) == 1
    assert users[0].username == 'my-user'

    # we need to explicitly cleanup because sub-process did not use
    # our db fixture, this will be done at teardown in the db_cleanup fixture



from .oauth2 import OAuthAuthenticator


class GitlabAuthenticator(OAuthAuthenticator):
    """Use Gitlab account to authenticate users with Quetz.

    To enable add the following to the configuration file:

    .. code::

      [gitlab]
      client_id = "fde330aef1fbe39991"
      client_secret = "03728444a12abff17e9444fd231b4379d58f0b"

    The above will use `<https://gitlab.com>`_. You can specify a self-hosted
    GitLab instance using the ``url`` parameter:

    .. code::

      [gitlab]
      url = "https://gitlab.mydomain.org"
      client_id = "fde330aef1fbe39991"
      client_secret = "03728444a12abff17e9444fd231b4379d58f0b"

    You can obtain ``client_id`` and ``client_secret`` by registering your
    application with Gitlab at this URL:
    `<https://gitlab.com/-/profile/applications>`_ or
    `<https://gitlab.mydomain.org/admin/applications>`_
    if using a self-hosted GitLab instance.
    Select ``openid`` as scope. If you want to collect email addresses,
    make sure to also select ``email``, ``profile`` and ``read_user`` as scope
    in the Gitlab interface.
    """

    provider = "gitlab"

    collect_emails = False

    # oauth client params
    scope = 'openid'

    # endpoint urls
    validate_token_url = "user"

    async def userinfo(self, request, token):
        resp = await self.client.get('/oauth/userinfo', token=token)
        profile = resp.json()
        gitlab_profile = {
            "id": profile["sub"],
            "name": profile["name"],
            "avatar_url": profile["picture"],
            "login": profile["nickname"],
        }

        # https://docs.gitlab.com/ee/integration/openid_connect_provider.html#shared-information
        if self.collect_emails:
            emails = await self.client.get('/api/v4/user/emails', token=token)
            emails_res = []
            emails_res.append(
                {
                    "email": profile["email"],
                    "primary": True,
                    "verified": profile["email_verified"],
                }
            )

            for e in emails.json():
                # there is a bug in gitlab so this should never be true (currently)
                if e["email"] == profile["email"]:
                    continue

                x = {
                    "email": e["email"],
                    "primary": False,
                    "verified": e["confirmed_at"] is not None,
                }
                emails_res.append(x)

            gitlab_profile["emails"] = emails_res

        return gitlab_profile

    def configure(self, config):
        if config.configured_section("gitlab"):
            self.access_token_url = f'{config.gitlab_url}/oauth/token'
            self.authorize_url = f'{config.gitlab_url}/oauth/authorize'
            self.api_base_url = f'{config.gitlab_url}/api/v4'
            self.revoke_url = f'{config.gitlab_url}/oauth/applications'
            self.client_id = config.gitlab_client_id
            self.client_secret = config.gitlab_client_secret
            self.is_enabled = True
            if config.configured_section("users"):
                self.collect_emails = config.users_collect_emails
                self.scope = 'openid email read_user'

        else:
            self.is_enabled = False

        # call the configure of base class to set default_channel and default role
        super().configure(config)



import pytest
from fastapi import Request, Response
from starlette.testclient import TestClient

from quetz.authentication.base import BaseAuthenticationHandlers, BaseAuthenticator


class DummyHandlers(BaseAuthenticationHandlers):
    authorize_methods = ["POST"]

    async def login(self, request: Request):
        return Response("success")


class DummyAuthenticator(BaseAuthenticator):
    handler_cls = DummyHandlers
    provider = 'testprovider'

    def configure(self, config):
        self.is_enabled = True

    async def authenticate(self, request, data, **kwargs):
        return "dummy-user"


@pytest.fixture
def dummy_authenticator(app, config):
    from quetz.main import auth_registry

    authenticator = DummyAuthenticator(config)

    auth_registry.register(authenticator)

    return authenticator


def test_login_endpoint(app, dummy_authenticator):
    client = TestClient(app)

    response = client.get(f"/auth/{dummy_authenticator.provider}/login")

    assert response.status_code == 200
    assert response.text == "success"



"""add task pending state

Revision ID: 30241b33d849
Revises: cd404ed93cc0
Create Date: 2021-01-07 14:39:43.251123

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = '30241b33d849'
down_revision = 'cd404ed93cc0'
branch_labels = None
depends_on = None


def upgrade():
    # manually entered

    if op.get_context().dialect.name == 'postgresql':
        # https://alembic.sqlalchemy.org/en/latest/api/runtime.html#alembic.runtime.migration.MigrationContext.autocommit_block
        with op.get_context().autocommit_block():
            op.execute("ALTER TYPE taskstatus  ADD VALUE 'created'")
    else:
        # sqlite uses varchar + constraint for enum types
        taskstatus_enum = sa.Enum(
            'created',
            'pending',
            'running',
            'success',
            'failed',
            'skipped',
            name='taskstatus',
        )

        with op.batch_alter_table("tasks") as batch_op:
            batch_op.alter_column("status", type_=taskstatus_enum)


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    pass
    # ### end Alembic commands ###



# Copyright 2020 QuantStack
# Distributed under the terms of the Modified BSD License.

import random
import uuid

from quetz.config import Config
from quetz.database import get_session, init_db
from quetz.db_models import (
    ApiKey,
    Channel,
    ChannelMember,
    Identity,
    Package,
    PackageMember,
    Profile,
    User,
)


def init_test_db():
    config = Config()
    init_db(config.sqlalchemy_database_url)
    db = get_session(config.sqlalchemy_database_url)

    testUsers = []

    try:
        for index, username in enumerate(['alice', 'bob', 'carol', 'dave']):
            user = User(id=uuid.uuid4().bytes, username=username)

            identity = Identity(
                provider='dummy',
                identity_id=str(index),
                username=username,
            )

            profile = Profile(name=username.capitalize(), avatar_url='/avatar.jpg')

            user.identities.append(identity)
            user.profile = profile
            db.add(user)
            testUsers.append(user)

        for channel_index in range(30):
            channel = Channel(
                name=f'channel{channel_index}',
                description=f'Description of channel{channel_index}',
                private=False,
            )

            for package_index in range(random.randint(5, 100)):
                package = Package(
                    name=f'package{package_index}',
                    description=f'Description of package{package_index}',
                )
                channel.packages.append(package)

                test_user = testUsers[random.randint(0, len(testUsers) - 1)]
                package_member = PackageMember(
                    package=package, channel=channel, user=test_user, role='owner'
                )

                db.add(package_member)

            if channel_index == 0:
                package = Package(name='xtensor', description='Description of xtensor')
                channel.packages.append(package)

                test_user = testUsers[random.randint(0, len(testUsers) - 1)]
                package_member = PackageMember(
                    package=package, channel=channel, user=test_user, role='owner'
                )

                db.add(package_member)

                # create API key
                key = 'E_KaBFstCKI9hTdPM7DQq56GglRHf2HW7tQtq6si370'

                key_user = User(id=uuid.uuid4().bytes)

                api_key = ApiKey(
                    key=key, description='test API key', user=key_user, owner=test_user
                )
                db.add(api_key)

                key_package_member = PackageMember(
                    user=key_user,
                    channel_name=channel.name,
                    package_name=package.name,
                    role='maintainer',
                )
                db.add(key_package_member)

            db.add(channel)

            channel_member = ChannelMember(
                channel=channel,
                user=testUsers[random.randint(0, len(testUsers) - 1)],
                role='owner',
            )

            db.add(channel_member)
        db.commit()
    finally:
        db.close()


if __name__ == '__main__':
    init_test_db()



"""create indexes for download count queries

Revision ID: b9886d9cadb0
Revises: 0653794b6252
Create Date: 2021-01-28 13:11:27.197409

"""
from alembic import op

# revision identifiers, used by Alembic.
revision = 'b9886d9cadb0'
down_revision = '0653794b6252'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('aggregated_metrics', schema=None) as batch_op:
        batch_op.create_index(
            'package_version_metric_index',
            [
                'channel_name',
                'platform',
                'filename',
                'metric_name',
                'period',
                'timestamp',
            ],
            unique=False,
        )
        batch_op.create_unique_constraint(
            "package_version_metric_constraint",
            [
                'channel_name',
                'platform',
                'filename',
                'metric_name',
                'period',
                'timestamp',
            ],
        )

    with op.batch_alter_table('package_versions', schema=None) as batch_op:
        batch_op.create_index(
            'package_version_filename_index',
            ['channel_name', 'filename', 'platform'],
            unique=True,
        )

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('package_versions', schema=None) as batch_op:
        batch_op.drop_index('package_version_filename_index')

    with op.batch_alter_table('aggregated_metrics', schema=None) as batch_op:
        batch_op.drop_constraint("package_version_metric_constraint", type_='unique')
        batch_op.drop_index('package_version_metric_index')

    # ### end Alembic commands ###



import logging
import os
import tempfile

import pytest

from quetz.config import Config, ConfigEntry, ConfigSection, configure_logger
from quetz.dao import Dao
from quetz.errors import ConfigError


@pytest.fixture
def config_extra():
    return r"""[users]
admins=["bartosz"]
default_role = "member"
create_default_channel = true
"""


def test_config_without_file_path_set(config_str):
    # the env variable should not be defined for this test to work
    assert not os.environ.get("QUETZ_CONFIG_FILE")

    # we need to check whether Config was not initialised before
    assert not Config._instances

    # check if it works with path even if QUETZ_CONFIG_FILE is
    # not defined
    with tempfile.NamedTemporaryFile("w", delete=False) as fid:
        fid.write(config_str)
        fid.flush()
        config = Config(fid.name)
    assert config.configured_section("users")


def test_config_users(config):
    assert config.users_default_role == "member"
    assert config.users_create_default_channel
    assert config.users_admins == ["bartosz"]
    assert not config.users_maintainers
    assert not config.users_members


@pytest.mark.parametrize(
    "config_extra", ["[users]\nadmins=[]", "[users]\nmaintainers=[]"]
)
def test_config_empty_users_section(dao: Dao, user, config):
    assert config.configured_section("users")
    assert not config.users_admins
    assert not config.users_maintainers
    assert not config.users_members
    assert not config.users_default_role
    assert not config.users_create_default_channel


def test_config_is_singleton(config):
    c = Config()

    assert c is config

    Config._instances = {}

    c_new = Config()

    assert c_new is not config

    c_file = Config("config.toml")

    assert c_file is c_new


def test_config_with_path(config_dir, config_base):
    one_path = os.path.join(config_dir, "one_config.toml")
    other_path = os.path.join(config_dir, "other_config.toml")
    with open(one_path, 'w') as fid:
        fid.write("\n".join([config_base, "[users]\nadmins=['one']"]))
    with open(other_path, 'w') as fid:
        fid.write("\n".join([config_base, "[users]\nadmins=['other']"]))

    Config._instances = {}

    c_one = Config(one_path)

    assert c_one.configured_section("users")
    assert c_one.users_admins == ["one"]

    c_other = Config(other_path)

    assert c_other.configured_section("users")
    assert c_other.users_admins == ["other"]

    c_new = Config(one_path)

    assert c_new is c_one


def test_config_extend_require(config):
    with pytest.raises(ConfigError):
        config.register(
            [
                ConfigSection(
                    "other_plugin",
                    [
                        ConfigEntry("some_config_value", str),
                    ],
                )
            ]
        )
    # remove last entry again
    config._config_map.pop()


@pytest.mark.parametrize(
    "config_extra", ["[extra_plugin]\nsome=\"testvalue\"\nconfig=\"othervalue\"\n"]
)
def test_config_extend(config):
    config.register(
        [
            ConfigSection(
                "extra_plugin",
                [
                    ConfigEntry("some", str),
                    ConfigEntry("config", str),
                    ConfigEntry("has_default", str, "iamdefault"),
                ],
            )
        ]
    )

    assert config.extra_plugin_some == 'testvalue'
    assert config.extra_plugin_config == 'othervalue'
    assert config.extra_plugin_has_default == 'iamdefault'

    config._config_map.pop()


def test_configure_logger(capsys):
    "configure_logger should be idempotent"

    configure_logger()
    logger = logging.getLogger("quetz")
    logger.error("my test")

    captured = capsys.readouterr()
    assert "[quetz]" in captured.err
    assert "ERROR" in captured.err
    assert "my test" in captured.err
    assert len(captured.err.splitlines()) == 1

    captured = capsys.readouterr()
    assert not captured.err

    configure_logger()
    logger.info("second")
    captured = capsys.readouterr()
    assert "[quetz]" in captured.err
    assert "INFO" in captured.err
    assert "second" in captured.err
    assert captured.err.count("second") == 1
    assert "my test" not in captured.err
    assert len(captured.err.splitlines()) == 1



import pytest

from quetz.dao import Dao
from quetz.db_models import ApiKey, ChannelMember, PackageMember
from quetz.rest_models import BaseApiKey
from quetz.utils import generate_random_key


@pytest.fixture
def api_keys(other_user, user, db, dao: Dao):
    def key_factory(key_user, descr, expire_at, roles):
        return dao.create_api_key(
            key_user.id,
            BaseApiKey.parse_obj(
                dict(description=descr, expire_at=expire_at, roles=roles)
            ),
            descr,
        )

    keys = [
        key_factory(user, "key", '2030-01-01', []),
        key_factory(other_user, "other_key", '2030-01-01', []),
        key_factory(other_user, "other_user_is_user", '2030-01-01', []),
        key_factory(user, "user_is_user", '2030-01-01', []),
    ]

    yield {k.description: k for k in keys}

    for key in keys:
        db.delete(key)
    db.commit()


@pytest.mark.parametrize(
    "user_role,target_key,expected_status",
    [
        ("owner", "other_key", 200),
        ("maintainer", "other_key", 200),
        ("member", "other_key", 403),
        (None, "other_key", 403),
        ("maintainer", "key", 200),
        ("member", "key", 200),
        (None, "key", 200),
        (None, "other_user_is_user", 403),
        (None, "user_is_user", 200),
    ],
)
def test_delete_api_key(auth_client, api_keys, db, target_key, expected_status):
    response = auth_client.delete(f"/api/api-keys/{target_key}")

    assert response.status_code == expected_status

    deleted_key = db.query(ApiKey).filter(ApiKey.key == target_key).one_or_none()

    if expected_status == 200:
        assert deleted_key.deleted
    else:
        assert not deleted_key.deleted


def test_delete_api_key_does_not_exist(auth_client):
    response = auth_client.delete("/api/api-keys/key")

    assert response.status_code == 404

    assert "does not exist" in response.json()["detail"]


def test_list_keys_with_channel_roles(
    auth_client, api_keys, db, user, other_user, private_channel
):
    channel_member = ChannelMember(
        channel=private_channel, user=user, role="maintainer"
    )
    db.add(channel_member)
    db.commit()

    response = auth_client.get("/api/api-keys")
    assert response.status_code == 200
    returned_keys = {key["description"]: key["roles"] for key in response.json()}
    assert len(returned_keys) == 2

    assert returned_keys["user_is_user"] is None
    assert returned_keys["key"] is None


def test_list_keys_with_package_roles(
    auth_client,
    api_keys,
    db,
    user,
    other_user,
    private_channel,
    private_package,
):
    package_member = PackageMember(
        channel=private_channel, package=private_package, user=user, role="maintainer"
    )
    db.add(package_member)
    db.commit()

    response = auth_client.get("/api/api-keys")
    assert response.status_code == 200
    returned_keys = {key["description"]: key["roles"] for key in response.json()}
    assert len(returned_keys) == 2

    assert returned_keys["user_is_user"] is None
    assert returned_keys["key"] is None


def test_list_keys_subrole(auth_client, dao, user, private_channel):
    dao.create_api_key(
        user.id,
        BaseApiKey.parse_obj(
            dict(
                description="user-key",
                roles=[
                    {"channel": private_channel.name, 'package': None, "role": "owner"}
                ],
            )
        ),
        "user-key",
    )

    response = auth_client.get("/api/api-keys")
    returned_keys = {key["description"]: key["roles"] for key in response.json()}
    assert len(returned_keys) == 1
    assert "user-key" in returned_keys
    assert returned_keys['user-key'] == [
        {"channel": private_channel.name, "package": None, "role": "owner"}
    ]


def test_list_keys_without_roles(auth_client, dao, user):
    dao.create_api_key(
        user.id,
        BaseApiKey.parse_obj(dict(description="user-key", roles=[])),
        "user-key",
    )

    response = auth_client.get("/api/api-keys")
    assert response.status_code == 200
    returned_keys = {key["description"]: key["roles"] for key in response.json()}
    assert len(returned_keys) == 1
    assert "user-key" in returned_keys
    assert returned_keys['user-key'] is None


def test_unlist_delete_api_keys(auth_client, api_keys, db, private_channel, user):
    channel_member = ChannelMember(
        channel=private_channel, user=user, role="maintainer"
    )
    db.add(channel_member)
    db.commit()

    response = auth_client.get("/api/api-keys")

    assert response.status_code == 200
    returned_keys = {key["description"]: key["roles"] for key in response.json()}
    assert "key" in returned_keys

    api_keys["key"].deleted = True
    db.commit()

    response = auth_client.get("/api/api-keys")
    assert response.status_code == 200

    returned_keys = {key["description"]: key["roles"] for key in response.json()}
    assert "key" not in returned_keys


@pytest.mark.parametrize("loop", [i + 1 for i in range(100)])
def test_generate_random_key(loop):
    key = generate_random_key()
    assert len(key) == 32
    assert key.isalnum()


@pytest.mark.parametrize("length", [i + 1 for i in range(100)])
def test_generate_random_key_variable_length(length):
    key = generate_random_key(length)
    assert len(key) == length
    assert key.isalnum()



import json
from datetime import datetime
from typing import Optional

from quetz.jobs.models import Job, JobStatus


class JobsDao:
    def __init__(self, db):
        self.db = db

    def create_job(
        self,
        job_manifest,
        user_id,
        extra_args={},
        start_at: Optional[datetime] = None,
        repeat_every_seconds: Optional[int] = None,
    ):
        extra_args_json: Optional[str]
        if extra_args:
            extra_args_json = json.dumps(extra_args)
        else:
            extra_args_json = None
        job = Job(
            manifest=job_manifest,
            owner_id=user_id,
            extra_args=extra_args_json,
            status=JobStatus.pending,
            start_at=start_at,
            repeat_every_seconds=repeat_every_seconds,
        )
        self.db.add(job)
        self.db.commit()
        return job



import os
import shutil
import tempfile
from typing import List

import pytest
from alembic.command import upgrade as alembic_upgrade
from fastapi.testclient import TestClient

import quetz
from quetz.cli import _alembic_config
from quetz.config import Config
from quetz.dao import Dao
from quetz.database import get_engine, get_session_maker
from quetz.db_models import Base


def pytest_configure(config):
    pytest.quetz_variables = {
        var: value for var, value in os.environ.items() if var.startswith("QUETZ_")
    }
    for var in pytest.quetz_variables:
        del os.environ[var]


def pytest_unconfigure(config):
    for var, value in pytest.quetz_variables.items():
        os.environ[var] = value


@pytest.fixture
def sqlite_in_memory():
    """whether to create a sqlite DB in memory or on the filesystem."""
    return True


@pytest.fixture
def sqlite_url(sqlite_in_memory):
    if sqlite_in_memory:
        yield "sqlite:///:memory:"
    else:
        try:
            sql_path = tempfile.TemporaryDirectory()
            yield f"sqlite:///{sql_path.name}/test_quetz.sqlite"
        finally:
            sql_path.cleanup()


@pytest.fixture
def database_url(sqlite_url):
    db_url = os.environ.get("QUETZ_TEST_DATABASE", sqlite_url)
    return db_url


@pytest.fixture
def sql_echo():
    """whether to activate SQL echo during the tests or not."""
    return False


@pytest.fixture
def engine(database_url, sql_echo):
    sql_echo = bool(os.environ.get("QUETZ_TEST_ECHO_SQL", sql_echo))
    engine = get_engine(database_url, echo=sql_echo, reuse_engine=False)
    yield engine
    engine.dispose()


@pytest.fixture
def use_migrations() -> bool:
    USE_MIGRATIONS = "use-migrations"
    CREATE_TABLES = "create-tables"
    migrations_env = os.environ.get("QUETZ_TEST_DBINIT", CREATE_TABLES)
    if migrations_env.lower() == CREATE_TABLES:
        return False
    elif migrations_env.lower() == USE_MIGRATIONS:
        return True
    else:
        raise ValueError(
            f"QUETZ_TEST_DBINIT should be either {CREATE_TABLES} or {USE_MIGRATIONS}"
        )


@pytest.fixture
def sql_connection(engine):
    connection = engine.connect()
    yield connection
    connection.close()


@pytest.fixture
def alembic_config(database_url, sql_connection):
    alembic_config = _alembic_config(database_url)
    alembic_config.attributes["connection"] = sql_connection
    return alembic_config


@pytest.fixture
def create_tables(alembic_config, engine, use_migrations):
    if use_migrations:
        alembic_upgrade(alembic_config, 'heads', sql=False)
    else:
        Base.metadata.create_all(engine)


@pytest.fixture
def auto_rollback():
    """Whether to revert automatically the changes in the database after each test.

    In most cases, you will want to set this flag to True, only sporadically you might
    run the tests outside a revertible transaction (for example, to test interaction
    between two concurrent clients). But then you will need to clean up the db objects
    manually after each test. Use with care.

    See also the comment in session_maker fixture."""

    return True


@pytest.fixture
def session_maker(sql_connection, create_tables, auto_rollback):
    # run the tests with a separate external DB transaction
    # so that we can easily rollback all db changes (even if commited)
    # done by the test client

    # Note: when rollback is explictly called in the implementation,
    #       it will remove all objects created in the test even the ones
    #       that were already committed!

    # see also: https://docs.sqlalchemy.org/en/13/orm/session_transaction.html#joining-a-session-into-an-external-transaction-such-as-for-test-suites # noqa

    if auto_rollback:
        trans = sql_connection.begin()

    sql_connection.name = 'sqlite-test'
    yield get_session_maker(sql_connection)

    if auto_rollback:
        trans.rollback()


@pytest.fixture
def expires_on_commit():
    return True


@pytest.fixture
def db(session_maker, expires_on_commit):
    session = session_maker()
    session.expire_on_commit = expires_on_commit
    yield session
    session.close()


@pytest.fixture
def config_auth():
    return """
[github]
# Register the app here: https://github.com/settings/applications/new
client_id = "aaa"
client_secret = "bbb"
"""


@pytest.fixture
def config_base(database_url, plugins, config_auth):
    return f"""
{config_auth}

[sqlalchemy]
database_url = "{database_url}"

[session]
secret = "eWrkA6xpa7LTSSYUwZEEVoOU62501Ucf9lmLcgzTj1I="
https_only = false

[plugins]
enabled = {plugins}
"""


@pytest.fixture
def config_extra():
    return ""


@pytest.fixture
def config_str(config_base, config_extra):
    return "\n".join([config_base, config_extra])


@pytest.fixture
def home():
    return os.path.abspath(os.path.curdir)


@pytest.fixture
def config_dir(home):
    path = tempfile.mkdtemp()
    yield path
    shutil.rmtree(path)


@pytest.fixture(scope="session")
def test_data_dir():
    return os.path.join(os.path.dirname(quetz.__file__), "tests", "data")


@pytest.fixture
def config(config_str, config_dir, test_data_dir):
    config_path = os.path.join(config_dir, "config.toml")
    with open(config_path, "w") as fid:
        fid.write(config_str)
    old_dir = os.path.abspath(os.curdir)
    os.chdir(config_dir)
    os.environ["QUETZ_CONFIG_FILE"] = config_path
    for filename in os.listdir(test_data_dir):
        full_path = os.path.join(test_data_dir, filename)
        dest = os.path.join(config_dir, filename)
        if os.path.isfile(full_path):
            shutil.copy(full_path, dest)

    Config._instances = {}
    config = Config()
    yield config
    if "QUETZ_CONFIG_FILE" in os.environ:
        del os.environ["QUETZ_CONFIG_FILE"]
    Config._instances = {}
    os.chdir(old_dir)


@pytest.fixture
def plugins() -> List[str]:
    return []


@pytest.fixture
def app(config, db, mocker):
    # frontend router catches all urls
    # and takes priority over routes added after (in tests)
    # so we avoid adding it here
    mocker.patch("quetz.frontend.register")

    from quetz.deps import get_db
    from quetz.main import app

    # mocking is required for some functions that do not use fastapi
    # dependency injection (mainly non-request functions)
    def get_session_mock(*args, **kwargs):
        return db

    mocker.patch("quetz.database.get_session", get_session_mock)

    # overiding dependency works with all requests handlers that
    # depend on quetz.deps.get_db
    app.dependency_overrides[get_db] = lambda: db

    # root url was removed with fronted urls above but
    # redirects to root must work for some tests to pass
    @app.get("/")
    def root_endpoint():
        return "root"

    yield app
    app.dependency_overrides.pop(get_db)


@pytest.fixture
def client(app):
    client = TestClient(app)
    return client


@pytest.fixture
def auth_client(client, user):
    """authenticated client"""
    response = client.get(f"/api/dummylogin/{user.username}")
    assert response.status_code == 200
    return client


@pytest.fixture
def dao(db) -> Dao:
    return Dao(db)



"""merge ebe550f9fbbe and b9886d9cadb0

Revision ID: ea6eba9a9ffc
Revises: ebe550f9fbbe, b9886d9cadb0
Create Date: 2021-02-09 16:40:52.637115

"""

# revision identifiers, used by Alembic.
revision = 'ea6eba9a9ffc'
down_revision = ('ebe550f9fbbe', 'b9886d9cadb0')
branch_labels = None
depends_on = None


def upgrade():
    pass


def downgrade():
    pass



import logging
import os
import sys

from fastapi import APIRouter, Depends, HTTPException, status
from fastapi.responses import FileResponse
from importlib_metadata import entry_points

from quetz import authorization
from quetz.config import Config
from quetz.dao import Dao
from quetz.deps import get_dao, get_rules, get_session

config = Config()

logger = logging.getLogger('quetz')

catchall_router = APIRouter()

mock_settings_dict = None
frontend_dir = ""
config_data: dict


def _under_frontend_dir(path):
    """
    Check that path is under frontend_dir

    NOTE: os.path.abspath may seem unnecessary, but os.path.commonpath does not
    appear to handle relative paths as you would expect:

    >>> commonpath([abspath('../quetz/quetz'), abspath('quetz')])
    '/home/username/quetz/quetz'
    >>> commonpath(['../quetz/quetz', 'quetz'])
    ''
    """
    path = os.path.abspath(os.path.join(frontend_dir, path))
    fdir = os.path.abspath(frontend_dir)
    return os.path.commonpath([path, fdir]) == fdir


@catchall_router.get('/{resource:path}', include_in_schema=False)
def static(
    resource: str,
    session: dict = Depends(get_session),
    dao: Dao = Depends(get_dao),
    auth: authorization.Rules = Depends(get_rules),
):
    is_api_or_auth = resource.startswith(('api/', 'auth/'))
    if is_api_or_auth:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND)

    if "." not in resource:
        logger.info(f"returning index.html for {resource}")
        return FileResponse(path=os.path.join(frontend_dir, "index.html"))
    else:
        if not _under_frontend_dir(resource):
            raise HTTPException(status_code=status.HTTP_404_NOT_FOUND)

        file = os.path.join(frontend_dir, resource)
        if os.path.exists(file):
            return FileResponse(path=file)
        else:
            raise HTTPException(status_code=status.HTTP_404_NOT_FOUND)


def register(app):
    frontend_plugins = []
    for entry_point in entry_points().select(group='quetz.frontend'):
        frontend_plugins.append(entry_point)

    if len(frontend_plugins) > 1:
        logger.warning(
            "Multiple frontend plugins found!"
            f"{', '.join([str(fp) for fp in frontend_plugins])}\n"
            "Using last found."
        )

    if frontend_plugins:
        print("Register frontend hooks: ", frontend_plugins)
        logger.info(f"Loading frontend plugin: {frontend_plugins[-1]}")
        frontend_plugin = frontend_plugins[-1].load()
        return frontend_plugin.register(app)

    global frontend_dir
    global config_data

    # TODO do not add this in the final env, use nginx to route to static files
    app.include_router(catchall_router)

    if hasattr(config, 'general_frontend_dir') and config.general_frontend_dir:
        frontend_dir = config.general_frontend_dir
        logger.info(f"Configured frontend found: {config.general_frontend_dir}")
    elif os.path.isfile(f"{sys.prefix}/share/quetz/frontend/index.html"):
        logger.info("installed frontend found")
        frontend_dir = f"{sys.prefix}/share/quetz/frontend/"
    else:
        logger.info("Using basic fallback frontend")
        frontend_dir = os.path.join(
            os.path.dirname(os.path.realpath(__file__)), "basic_frontend"
        )



# Copyright 2020 QuantStack
# Distributed under the terms of the Modified BSD License.

from typing import List, Optional

from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm import Session

from quetz import rest_models
from quetz.authorization import OWNER
from quetz.config import Config
from quetz.dao import Dao
from quetz.db_models import Channel, Email, Identity, User
from quetz.errors import ValidationError

from . import base


def create_user_with_identity(
    dao: Dao,
    provider: str,
    profile: 'base.UserProfile',
    default_role: Optional[str],
    default_channels: Optional[List[str]],
) -> User:
    username = profile["login"]
    user = dao.create_user_with_profile(
        username=username,
        provider=provider,
        emails=profile.get("emails", []),
        identity_id=profile["id"],
        name=profile["name"],
        avatar_url=profile["avatar_url"],
        role=default_role,
        exist_ok=False,
    )

    if default_channels is not None:
        for channel_name in default_channels:
            i = 0

            while (
                dao.db.query(Channel).filter(Channel.name == channel_name).one_or_none()
            ):
                channel_name = f"{username}-{i}"

                i += 1

            channel_meta = rest_models.Channel(
                name=channel_name,
                description=f"{username}'s default channel",
                private=True,
            )

            dao.create_channel(channel_meta, user.id, OWNER)

    return user


def user_profile_changed(user, identity, profile: 'base.UserProfile'):
    if (
        identity.username != profile['login']
        or user.profile.name != profile['name']
        or user.profile.avatar_url != profile['avatar_url']
        or set((e.verified, e.primary, e.email) for e in user.emails)
        != set(
            (e['verified'], e['primary'], e['email']) for e in profile.get('emails', [])
        )
    ):
        return True

    return False


def update_user_from_profile(
    db: Session, user, identity, profile: 'base.UserProfile'
) -> User:
    identity.username = profile['login']
    user.profile.name = profile['name']
    user.profile.avatar_url = profile['avatar_url']

    # check if any email already registered
    emails = []
    for e in profile.get('emails', []):
        if not e["verified"]:
            continue

        user_email = db.query(Email).filter(Email.email == e["email"]).one_or_none()
        if user_email and user_email.user_id != user.id:
            raise IntegrityError(
                f"User {user.profile.name} already registered"
                f" with email {user_email.email}",
                "",
                "",
            )

        if user_email and e["verified"]:
            emails.append(user_email)
            user_email.primary = True if e["primary"] else False
        else:
            emails.append(
                Email(
                    email=e["email"],
                    verified=e["verified"],
                    primary=e["primary"],
                    provider=identity.provider,
                    identity_id=identity.identity_id,
                )
            )

    user.emails = emails

    db.commit()
    db.refresh(user)
    return user


def get_user_by_identity(
    dao: Dao,
    provider: str,
    profile: 'base.UserProfile',
    config: Config,
    default_role: Optional[str] = None,
    default_channels: Optional[List[str]] = None,
) -> User:
    db = dao.db

    try:
        user, identity = db.query(User, Identity).join(Identity).filter(
            Identity.provider == provider
        ).filter(
            Identity.identity_id == str(profile['id']),
        ).one_or_none() or (
            None,
            None,
        )
    except KeyError:
        print(f"unexpected response format: {profile}")

    if user:
        if user_profile_changed(user, identity, profile):
            return update_user_from_profile(db, user, identity, profile)
        return user

    try:
        user = create_user_with_identity(
            dao, provider, profile, default_role, default_channels
        )
    except IntegrityError:
        raise ValidationError(f"user name '{profile['login']}' already exists")

    return user



# Copyright 2020 Codethink Ltd
# Distributed under the terms of the Modified BSD License.

import json

from quetz.versionorder import VersionOrder

CHANNELDATA_OPTIONAL_FIELDS = (
    "description",
    "dev_url",
    "doc_source_url",
    "doc_url",
    "home",
    "icon_hash",
    "icon_url",
    "license",
    "license_family",
    "spdx_license",
    "source_git_url",
    "reference_package",
    "source_url",
    "summary",
    "version",
    "plugin_metadata",
)
CHANNELDATA_BINARY_FIELDS = (
    "activate.d",
    "deactivate.d",
    "post_link",
    "pre_link",
    "pre_unlink",
    "binary_prefix",
    "text_prefix",
)


def combine(old_data, new_data):
    if old_data is None:
        data = new_data
    else:
        data = {}
        newer = VersionOrder(old_data.get("version", "0")) < VersionOrder(
            new_data.get("version", "0")
        )
        for field in CHANNELDATA_BINARY_FIELDS:
            data[field] = any((new_data.get(field, False), old_data.get(field, False)))

        for field in ("keywords", "identifiers", "tags"):
            if newer and new_data.get(field):
                data[field] = new_data[field]
            else:
                data[field] = old_data.get(field, {})

        for field in CHANNELDATA_OPTIONAL_FIELDS:
            if newer and field in new_data:
                data[field] = new_data[field]
            elif field in old_data:
                data[field] = old_data[field]

        run_exports = old_data.get("run_exports", {})
        if "run_exports" in new_data:
            if new_data["run_exports"]:
                run_exports[new_data["version"]] = new_data["run_exports"]
        data["run_exports"] = run_exports

        data["timestamp"] = max(
            old_data.get("timestamp", 0), new_data.get("timestamp", 0)
        )

        data["subdirs"] = sorted(
            list(set(new_data.get("subdirs", [])) | set(old_data.get("subdirs", [])))
        )

    data = dict(sorted(data.items(), key=lambda item: item[0]))

    return data


def export(dao, channel_name):
    channeldata = {"channeldata_version": 1, "packages": {}, "subdirs": {}}
    packages = channeldata["packages"]
    subdirs = set(['noarch'])

    for name, info in dao.get_channel_datas(channel_name):
        if info is not None:
            data = json.loads(info)
            packages[name] = data
            subdirs = set(data.get("subdirs", [])) | subdirs

    channeldata["subdirs"] = list(subdirs)

    return channeldata



from datetime import datetime
from typing import Optional

from fastapi import APIRouter, Depends, HTTPException, status

from quetz import db_models
from quetz.dao import Dao
from quetz.deps import get_dao, get_package_or_fail
from quetz.metrics import rest_models
from quetz.metrics.db_models import IntervalType

api_router = APIRouter(prefix="/metrics")


@api_router.get(
    "/channels/{channel_name}/packages/{package_name}/versions/{platform}/{filename}",  # noqa
    response_model=rest_models.PackageVersionMetricResponse,
    tags=["metrics"],
)
def get_package_version_metrics(
    platform: str,
    filename: str,
    package_name: str,
    channel_name: str,
    fill_zeros: bool = False,
    period: IntervalType = IntervalType.day,
    metric_name: str = "download",
    start: Optional[datetime] = None,
    end: Optional[datetime] = None,
    package: db_models.Package = Depends(get_package_or_fail),
    dao: Dao = Depends(get_dao),
):
    version = dao.get_package_version_by_filename(
        channel_name, package_name, filename, platform
    )

    if not version:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=f"package version {platform}/{filename} not found",
        )

    series = dao.get_package_version_metrics(
        version.id, period, metric_name, start=start, end=end, fill_zeros=fill_zeros
    )

    total = sum(s.count for s in series)

    return {
        "server_timestamp": datetime.utcnow(),
        "period": period,
        "metric_name": metric_name,
        "total": total,
        "series": series,
    }


@api_router.get(
    "/channels/{channel_name}",
    response_model=rest_models.ChannelMetricResponse,
    tags=["metrics"],
)
def get_channel_metrics(
    channel_name: str,
    period: IntervalType = IntervalType.day,
    metric_name: str = "download",
    platform: Optional[str] = None,
    start: Optional[datetime] = None,
    end: Optional[datetime] = None,
    dao: Dao = Depends(get_dao),
):
    metrics = dao.get_channel_metrics(
        channel_name, period, metric_name, platform=platform, start=start, end=end
    )

    return {
        "server_timestamp": datetime.utcnow(),
        "period": period,
        "metric_name": metric_name,
        "packages": metrics,
    }


def get_router():
    return api_router



# Copyright 2020 QuantStack
# Distributed under the terms of the Modified BSD License.
import logging
from contextlib import contextmanager
from typing import Callable

from sqlalchemy import create_engine, event
from sqlalchemy.engine import Engine
from sqlalchemy.orm import sessionmaker
from sqlalchemy.orm.session import Session
from sqlalchemy.pool import StaticPool

from quetz.config import Config
from quetz.metrics.middleware import DATABASE_CONNECTIONS_USED, DATABASE_POOL_SIZE

engine = None

logger = logging.getLogger("quetz")


def set_metrics(*args):
    checked_in = engine.pool.checkedin()
    checked_out = engine.pool.checkedout()
    pool_size = checked_in + checked_out
    DATABASE_POOL_SIZE.set(pool_size)
    DATABASE_CONNECTIONS_USED.set(checked_out)


def get_engine(db_url, echo: bool = False, reuse_engine=True, **kwargs) -> Engine:
    if db_url.startswith('sqlite'):
        kwargs.setdefault('connect_args', {'check_same_thread': False})

    if db_url.endswith(':memory:'):
        # If we're using an in-memory database, ensure that only one connection
        # is ever created.
        kwargs.setdefault('poolclass', StaticPool)

    global engine

    if not engine or not reuse_engine:
        # TODO make configurable!
        if db_url.startswith('postgres'):
            engine = create_engine(
                db_url, echo=echo, pool_size=200, max_overflow=100, **kwargs
            )
            for event_name in ['connect', 'close', 'checkin', 'checkout']:
                event.listen(engine, event_name, set_metrics)
        else:
            engine = create_engine(db_url, echo=echo, **kwargs)

        def on_connect(dbapi_conn, conn_record):
            logger.debug("connection opened: %s", engine.pool.status())

        def on_close(dbapi_conn, conn_record):
            logger.debug("connection closed: %s", engine.pool.status())

        event.listen(engine, 'connect', on_connect)
        event.listen(engine, 'close', on_close)

    return engine


def get_session_maker(engine) -> Callable[[], Session]:
    return sessionmaker(autocommit=False, autoflush=True, bind=engine)


def get_session(db_url: str, echo: bool = False, **kwargs) -> Session:
    """Get a database session.

    Important note: this function is mocked during tests!

    """
    return get_session_maker(get_engine(db_url, echo, **kwargs))()


@contextmanager
def get_db_manager():
    config = Config()
    database_url = config.sqlalchemy_database_url
    db = get_session(database_url, echo=config.sqlalchemy_echo_sql)

    try:
        yield db
    finally:
        db.close()
